/***************************************************************************
                          array3ditem.h  -  description
                             -------------------
    begin                : Tue Dec 25 2001
    copyright            : (C) 2015 by Werner Stille
    email                : stille@uni-freiburg.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef ARRAY3DITEM_H
#define ARRAY3DITEM_H

#include <QTransform>
#include "arrayitem.h"
#include "threeditem.h"

/**
  * 3D arrayitem class. Provides methods for 3D plotting of arrays.
  * @author Werner Stille
  */
class KDE_EXPORT Array3DItem : public ThreeDItem
{
public:
  /** Specific QGraphicsItem type number for function type() */
  enum {
    Type = UserType + Array3D
    /**< Specific QGraphicsItem type number of Array3DItem for function
      * type().
      */
  };
  /**
    * Constructor. Initializes corresponding to arguments.
    * @param x0 left page margin in centimeters.
    * @param y0 bottom page margin in centimeters.
    * @param width width in centimeters.
    * @param height height in centimeters.
    * @param colorFrame color for frame and labels.
    * @param colorGrid color for grids and tics.
    * @param colorData color for data plot.
    * @param logX true for logarithmic x axis.
    * @param logY true for logarithmic y axis.
    * @param logZ true for logarithmic y axis.
    * @param gridMode 0 for frame only, 1 for axes with tics,
    * 2 for axes with tics and labels.
    * @param mode3D 3D mode.
    * @param xMax maximum x value.
    * @param yMax maximum y value.
    * @param zMax maximum z value.
    * @param textX x axis text.
    * @param textY y axis text.
    * @param textZ z axis text.
    * @param header heading text.
    * @param relSize relative size of text, markers and lines.
    * @param colX x column index.
    * @param colY y column index.
    * @param colZ z column index.
    * @param colErr y error column index.
    * @param errorBars true to plot error bars.
    * @param decimalChar decimal symbol of floating point numbers used in file.
    * @param numDigX number of digits for x axis.
    * @param numDigY number of digits for y axis.
    * @param numDigZ number of digits for z axis.
    * @param minTicX number of intervals between major tics of x axis.
    * @param minTicY number of intervals between major tics of y axis.
    * @param minTicZ number of intervals between major tics of z axis.
    * @param expX logarithm of normalization factor for x axis.
    * @param expY logarithm of normalization factor for y axis.
    * @param expZ logarithm of normalization factor for z axis.
    * @param xMin minimum x value.
    * @param yMin minimum y value.
    * @param zMin minimum z value.
    * @param xTic distance between major tics of x axis.
    * @param yTic distance between major tics of y axis.
    * @param zTic distance between major tics of z axis.
    * @param deltaX x step size.
    * @param deltaZ z step size.
    * @param phi angle of rotation about y axis in deg.
    * @param theta tilt angle of y axis in deg.
    * @param boxX x box width in cm.
    * @param boxY y box width in cm.
    * @param boxZ z box width in cm.
    * @param xRef x position of center of box bottom plane.
    * @param yRef y position of center of box bottom plane.
    * @param fx normalization factor for x values.
    * @param fy normalization factor for y values.
    * @param fz normalization factor for z values.
    * @param rowStart start row index.
    * @param n number of data points.
    * @param smooth smoothing factor.
    * @param url URL of data file.
    * @param showFrame true to show frame.
    * @param internal true for internal stored array data.
    * @param collapsed true for collapsed in list view.
    * @param visible true for visible item.
    */
  Array3DItem(double x0 = 4, double y0 = 3, double width = 15,
              double height = 10, unsigned colorFrame = 0,
              unsigned colorGrid = 0, unsigned colorData = 0,
              bool logX = false, bool logY = false, bool logZ = false,
              GridMode gridMode = AxesWithLabels, Mode3D mode3D = Grid,
              double xMax = 15, double yMax = 10, double zMax = 15,
              const QString& textX = "", const QString& textY = "",
              const QString& textZ = "", const QString& header = "",
              double relSize = 1, int colX = 0, int colY = 1, int colZ = 2,
              int colErr = 0, bool errorBars = false,
              ArrayItem::DecimalCharacter decimalChar = ArrayItem::Point,
              int numDigX = -1, int numDigY = -1, int numDigZ = -1,
              int minTicX = 5, int minTicY = 2, int minTicZ = 5, int expX = 0,
              int expY = 0, int expZ = 0, double xMin = 0, double yMin = 0,
              double zMin = 0, double xTic = 5, double yTic = 2,
              double zTic = 5, double deltaX = 0, double deltaZ = 0,
              double phi = 20, double theta = -20, double boxX = 7.5,
              double boxY = 7.5, double boxZ = 7.5, double xRef = 7.5,
              double yRef = 4, double fx = 1, double fy = 1, double fz = 1,
              int rowStart = 0, int n = 0, double smooth = 0,
              const KUrl& url = KUrl(), bool showFrame = true,
              bool internal = false, bool collapsed = false,
              bool visible = true);
  /** Copy constructor. */
  Array3DItem(const Array3DItem& a);
  /**
    * Constructor. Initializes corresponding to plot file.
    * @param plo pointer to KConfigGroup object containing item data.
    * @param urlPlo URL of plot file.
    * @param x0 default value for left page margin in centimeters.
    * @param y0 default value for bottom page margin in centimeters.
    * @param width default value for width in centimeters.
    * @param height default value for height in centimeters.
    * @param colorFrame default value for color for frame and labels.
    * @param colorGrid default value for color for grids and tics.
    * @param colorData color for data plot.
    * @param gridMode default grid mode.
    * @param colX default x column index.
    * @param colY default y column index.
    * @param colErr default y error column index.
    * @param errorBars default, true to plot error bars.
    */
  Array3DItem(const KConfigGroup* plo, const KUrl& urlPlo, double x0 = 4,
              double y0 = 3, double width = 15, double height = 10,
              unsigned colorFrame = 0, unsigned colorGrid = 0,
              unsigned colorData = 0, GridMode gridMode = AxesWithLabels,
              int colX = 0, int colY = 1, int colErr = 0,
              bool errorBars = false);
  /** Destructor. */
  ~Array3DItem();
  /** Calculates value table. */
  void calcTable() const;
  /** Returns y error column index. */
  inline int colErr() const { return m_colErr; };
  /** Returns x column index. */
  inline int colX() const { return m_colX; };
  /** Returns y column index. */
  inline int colY() const { return m_colY; };
  /** Returns z column index. */
  inline int colZ() const { return m_colZ; };
  /** Returns number of data points. */
  inline int count() const { return m_count; };
  /**
    * Returns pointer to internal array.
    * @return pointer to two-dimensional array.
    */
  inline double*** data() { return &m_data; };
  /** Returns decimal symbol of floating point numbers.  */
  inline ArrayItem::DecimalCharacter decimalCharacter() const
  { return m_decimalCharacter; };
  /** Returns true to use error bars. */
  inline bool errorBars() const { return m_errorBars; };
  /** Returns true for internal stored array data. */
  inline bool internal() const { return m_internal; };
  /** Returns number of columns. */
  inline int numCols() const { return m_numCols; };
  /** Returns number of rows. */
  inline int numRows() const { return m_numRows; };
  /** Paints item.  */
  virtual void paint(QPainter* painter, const QStyleOptionGraphicsItem* option,
                     QWidget* widget = 0);
  /** Returns start row index. */
  inline int rowStart() const { return m_rowStart; };
  /**
    * Sets error column index.
    * @param colErr y error column index.
    */
  inline void setColErr(int colErr) { m_colErr = colErr; };
  /**
    * Sets x column index.
    * @param colX y error column index.
    */
  inline void setColX(int colX) { m_colX = colX; };
  /**
    * Sets y column index.
    * @param colY y column index.
    */
  inline void setColY(int colY) { m_colY = colY; };
  /**
    * Sets z column index.
    * @param colZ z column index.
    */
  inline void setColZ(int colZ) { m_colZ = colZ; };
  /**
    * Sets number of data points.
    * @param n number of data points.
    */
  inline void setCount(int n) { m_count = n; };
  /**
    * Sets decimal character.
    * @param decimalCharacter decimal symbol of floating point numbers used in
    * file.
    */
  inline void setDecimalCharacter(ArrayItem::DecimalCharacter decimalCharacter)
  { m_decimalCharacter = decimalCharacter; };
  /**
    * Sets error bars mode.
    * @param errorBars true to use error bars.
    */
  inline void setErrorBars(bool errorBars) { m_errorBars = errorBars; };
  /**
    * Sets storage mode.
    * @param internal true for internal stored array data.
    */
  inline void setInternal(bool internal) { m_internal = internal; };
  /**
    * Sets number of columns.
    * @param nCols number of columns.
    */
  inline void setNumCols(int nCols) { m_numCols = nCols; };
  /**
    * Sets number of rows.
    * @param nRows number of Rows.
    */
  inline void setNumRows(int nRows) { m_numRows = nRows; };
  /**
    * Sets start row index.
    * @param rowStart start row index.
    */
  inline void setRowStart(int rowStart) { m_rowStart = rowStart; };
  /**
    * Sets smoothing factor.
    * @param smooth factor.
    */
  inline void setSmf(double smooth) { m_smf = smooth; };
  /**
    * Sets URL.
    * @param url URL of data file.
    */
  inline void setUrl(const KUrl& url) { m_url = url; };
  /** Returns smoothing factor. */
  inline double smf() const { return m_smf; };
  /** Returns specific QGraphicsItem type number. */
  virtual int type() const { return Type; };
  /** Returns URL of data file. */
  inline const KUrl& url() const { return m_url; };
  /** Applies changes and forces redrawing. */
  void update();
  /**
   * Writes into plot file.
   * @param plo KConfigGroup object.
   * @param urlPlo URL of plot file.
   * @param absolute true to save absolute paths.
   */
  void writePlo(KConfigGroup* plo, const KUrl& urlPlo, bool absolute) const;

protected:
  bool m_internal, m_errorBars;
  mutable bool errorBarso;
  int m_colX, m_colY, m_colZ, m_colErr, m_rowStart, m_count, m_numRows,
      m_numCols;
  ArrayItem::DecimalCharacter m_decimalCharacter;
  mutable int colXo, colYo, colZo, colErro, rowStarto, counto;
  double m_smf;
  mutable double smfo;
  double** m_data;
  KUrl m_url;
};

#endif
