/***************************************************************************
                          frameitem.cpp  -  description
                             -------------------

    This file is a part of kpl - a program for graphical presentation of
    data sets and functions.

    begin                : Sun Aug 29 1999
    copyright            : (C) 2015 by Werner Stille
    email                : stille@uni-freiburg.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <QAbstractTextDocumentLayout>
#include <QPainter>
#include <QTextFrame>
#include <KConfigGroup>
#include "arrayitem.h"
#include "frameitem.h"
#include "kplscene.h"
#include "parfunitem.h"
#include "splineitem.h"
#include "textitem.h"

/**
 * Calculates normalization factors.
 * @param item pointer to graphics item.
 * @param iex reference to logarithm of x normalization factor.
 * @param iey reference to logarithm of x normalization factor.
 * @param fx reference to x normalization factor.
 * @param fy reference to y normalization factor.
 * @param logX true for logarithmic x axis.
 */
static void exponent(QGraphicsItem* item, int& iex, int& iey, double& fx,
                     double& fy, bool logX)
{
  switch (item->type()) {
    case ArrayItem::Type:
      qgraphicsitem_cast<ArrayItem*>(item)->exponent(iex, iey, fx, fy);
      break;
    case FunItem::Type:
      qgraphicsitem_cast<FunItem*>(item)->exponent(iex, iey, fx, fy, logX);
      break;
    case ParFunItem::Type:
      qgraphicsitem_cast<ParFunItem*>(item)->exponent(iex, iey, fx, fy);
      break;
    case SplineItem::Type:
      qgraphicsitem_cast<SplineItem*>(item)->exponent(iex, iey, fx, fy, logX);
  }
}

/**
 * Calculates minimum and maximum values.
 * @param item pointer to graphics item.
 * @param xMin reference to minimum of x values.
 * @param xMax reference to maximum of x values.
 * @param yMin reference to minimum of y values.
 * @param yMax reference to maximum of y values.
 * @param logX true for logarithmic x axis.
 */
static void limits(QGraphicsItem* item, double& xMin, double& xMax,
                   double& yMin, double& yMax, bool logX)
{
  switch (item->type()) {
    case ArrayItem::Type:
      qgraphicsitem_cast<ArrayItem*>(item)->limits(xMin, xMax, yMin, yMax);
      break;
    case FunItem::Type:
      qgraphicsitem_cast<FunItem*>(item)->limits(xMin, xMax, yMin, yMax, logX);
      break;
    case ParFunItem::Type:
      qgraphicsitem_cast<ParFunItem*>(item)->limits(xMin, xMax, yMin, yMax);
      break;
    case SplineItem::Type:
      qgraphicsitem_cast<SplineItem*>(item)->limits(xMin, xMax, yMin, yMax,
                                                    logX);
  }
}

FrameItem::FrameItem(double x0, double y0, double width, double height,
                     unsigned colorFrame, unsigned colorGrid, bool logX,
                     bool logY, GridMode gridMode, double xMax, double yMax,
                     const QString& textX, const QString& textY,
                     const QString& header, double relSize,
                     int numDigX, int numDigY, int minTicX, int minTicY,
                     int expX, int expY, double xMin, double yMin, double xTic,
                     double yTic, bool collapsed, bool visible, double xOffset,
                     double yOffset, bool xBottom, bool xTop, bool yLeft,
                     bool yRight) :
 KplItem(relSize), m_logX(logX), m_logY(logY), m_xBottom(xBottom),
 m_xTop(xTop), m_yLeft(yLeft), m_yRight(yRight), m_collapsed(collapsed),
 m_numDigX(numDigX), m_numDigY(numDigY), m_minTicX(minTicX),
 m_minTicY(minTicY), m_gridMode(gridMode), m_colorFrame(colorFrame),
 m_colorGrid(colorGrid), m_x0(x0), m_y0(y0), m_width(width), m_height(height),
 m_xMin(xMin), m_xMax(xMax), m_yMin(yMin), m_yMax(yMax), m_xTic(xTic),
 m_yTic(yTic), m_xOffset(xOffset), m_yOffset(yOffset), m_textX(textX),
 m_textY(textY), m_header(header)
{
  setExpX(expX);
  setExpY(expY);
  if (!visible)
    hide();
  setFlags(QGraphicsItem::ItemIsMovable | QGraphicsItem::ItemIsSelectable);
}

FrameItem::FrameItem(const FrameItem& a) :
 KplItem(a), QGraphicsRectItem(), m_logX(a.m_logX), m_logY(a.m_logY),
 m_xBottom(a.m_xBottom), m_xTop(a.m_xTop), m_yLeft(a.m_yLeft),
 m_yRight(a.m_yRight), m_collapsed(a.m_collapsed), m_numDigX(a.m_numDigX),
 m_numDigY(a.m_numDigY), m_minTicX(a.m_minTicX), m_minTicY(a.m_minTicY),
 m_expX(a.m_expX), m_expY(a.m_expY), m_gridMode(a.m_gridMode),
 m_colorFrame(a.m_colorFrame), m_colorGrid(a.m_colorGrid), m_x0(a.m_x0), m_y0(a.m_y0),
 m_width(a.m_width), m_height(a.m_height),  m_xMin(a.m_xMin), m_xMax(a.m_xMax),
 m_yMin(a.m_yMin), m_yMax(a.m_yMax), m_xTic(a.m_xTic), m_yTic(a.m_yTic),
 m_xOffset(a.m_xOffset), m_yOffset(a.m_yOffset), m_normX(a.m_normX), m_normY(a.m_normY),
 xminl(a.xminl), xmaxl(a.xmaxl), yminl(a.yminl), ymaxl(a.ymaxl),
 m_textX(a.m_textX), m_textY(a.m_textY), m_header(a.m_header), bRect(a.bRect),
 cRect(a.cRect), md(a.md), m_inv(a.m_inv), m_matrix(a.m_matrix)
{
  setZValue(a.zValue());
  setFlags(QGraphicsItem::ItemIsMovable | QGraphicsItem::ItemIsSelectable);
  setVisible(a.isVisible());
}

FrameItem::FrameItem(const KConfigGroup* plo, double x0, double y0,
                     double width, double height, unsigned colorFrame,
                     unsigned colorGrid, GridMode gridMode) :
 KplItem(plo)
{
  if (!plo->readEntry("active", true))
    hide();
  m_logX = plo->readEntry("logx", false);
  m_logY = plo->readEntry("logy", false);
  m_xTop = plo->readEntry("xTop", false);
  m_xBottom = plo->readEntry("xBottom", true);
  m_yRight = plo->readEntry("yRight", false);
  m_yLeft = plo->readEntry("yLeft", true);
  m_collapsed = plo->readEntry("collapsed", false);
  m_numDigX = plo->readEntry("ndigx", -1);
  m_numDigY = plo->readEntry("ndigy", -1);
  m_minTicX = plo->readEntry("mticx", 5);
  m_minTicY = plo->readEntry("mticy", 2);
  m_gridMode = GridMode(plo->readEntry("gridmode", int(gridMode)));
  setExpX(plo->readEntry("iex", 0));
  setExpY(plo->readEntry("iey", 0));
  QString s = plo->readEntry("colframe", "");
  m_colorFrame = s.isEmpty() ? colorFrame : rgb(s);
  s = plo->readEntry("colgrid", "");
  m_colorGrid = s.isEmpty() ? colorGrid : KplItem::rgb(s);
  m_x0 = plo->readEntry("xw0", x0);
  m_y0 = plo->readEntry("yw0", y0);
  m_width = plo->readEntry("xw", width);
  m_height = plo->readEntry("yw", height);
  m_xMin = plo->readEntry("xmi", 0.0);
  m_xMax = plo->readEntry("xma", m_xMin + 15);
  m_yMin = plo->readEntry("ymi", 0.0);
  m_yMax = plo->readEntry("yma", m_yMin + 10);
  m_xTic = plo->readEntry("xtic", 5.0);
  m_yTic = plo->readEntry("ytic", 2.0);
  m_xOffset = plo->readEntry("xOff", 0.0);
  m_yOffset = plo->readEntry("yOff", 0.0);
  m_textX = plo->readEntry("sx", "");
  m_textY = plo->readEntry("sy", "");
  m_header = plo->readEntry("sh", "");
  setFlags(QGraphicsItem::ItemIsMovable | QGraphicsItem::ItemIsSelectable);
}

FrameItem::~FrameItem()
{
}

void FrameItem::autoScale(double& xMin, double& xMax, double& tick, int& minTic,
                          int& numDig, double xmi, double xma, double f,
                          double c, bool loga)
{
  if (loga) {
    xmi = qMax(xmi, 1e-300);
    xma = qMax(xma, xmi);
    if (xmi == xma) {
      xmi = qMax(0.999 * xmi, 1e-300);
      xma = 1.001 * xmi;
    }
    numDig = -1;
    double lxmin = log10(xmi);
    double lxmax = log10(xma);
    double d = c * (lxmax - lxmin);
    int itic = 10;
    if (d > 1 || (qMax(fabs(lxmin), fabs(lxmax)) > 5)) {
      minTic = 1;
      int f = qMax(1, qRound(pow(10, floor(log10(d)))));
      d /= f;
      if (d > 7.071)
        itic = 10 * f;
      else
        if (d > 3.162)
          itic = 5 * f;
        else
          itic = ((d > 1.414) ? 2 : 1) * f;
      itic = qMin(itic, 100);
      xMin = pow(10, itic * floor(lxmin / itic + 1e-4));
      xMax = pow(10, itic * ceil(lxmax / itic + 1e-4));
      tick = pow(10, itic);
      return;
    } else {
      minTic = 10;
      if (d < 0.7)
        itic = (d < 0.5) ? 2 : 3;
      tick = itic;
    }
    double mant = lxmin - int(lxmin + 1e-6 * fabs(lxmin));
    if (fabs(mant) < 1e-9)
      mant = 0;
    if (mant < 0)
      ++mant;
    int e = qRound(lxmin - mant);
    if (itic == 2)
      if (mant < log10(1.999999))
        mant = 0;
      else
        if (mant < log10(4.999999))
          mant = log10(2);
        else
          mant = log10(5);
    else
      if (itic == 3)
        if (mant < log10(2.999999))
          mant = 0;
        else
          mant = log10(3);
      else
        mant = 0;
    xMin = pow(10, e + mant);
    mant = lxmax - int(lxmax + 1e-6 * fabs(lxmax));
    if (fabs(mant) < 1e-9)
      mant = 0;
    if (mant < 0)
      ++mant;
    e = qRound(lxmax - mant);
    if (itic == 2)
      if (mant > log10(5.000001))
        mant = 1;
      else
        if (mant > log10(2.000001))
          mant = log10(5);
        else {
          if (mant)
            mant = log10(2);
        }
    else
      if (itic == 3) {
        if (mant > log10(3.000001))
          mant = 1;
        else
          if (mant)
            mant = log10(3);
      } else
        if (mant)
          mant = 1;
    xMax = pow(10, e + mant);
  } else {
    if (xmi == xma) {
      double dx = xmi ? (1e-3 * fabs(xmi)) : (1e-3 / f);
      xmi -= dx;
      xma += dx;
    }
    double d = c * f * (xma - xmi);
    double f2 = pow(10, (-floor(log10(d + 1e-301))));
    d *= f2;
    if (d > 7.071)
      minTic = 10;
    else
      if (d > 3.162)
        minTic = 5;
      else
        minTic = (d > 1.414) ? 2 : 1;
    tick = minTic / f2;
    if ((minTic == 1) || (minTic == 10))
      minTic = 2;
    if (!(numDig = -int(floor(log10(tick + 1e-301) + 1e-6))))
      numDig = -1;
    xMin = tick * floor(f * xmi / tick + 1e-4);
    xMax = tick * floor(f * xma / tick + 0.9999);
  }
}

void FrameItem::autoScale(double xmi, double xma, double fx, double ymi,
                          double yma, double fy)
{
  autoScale(m_xMin, m_xMax, m_xTic, m_minTicX, m_numDigX, xmi, xma, fx, 0.2,
            m_logX);
  autoScale(m_yMin, m_yMax, m_yTic, m_minTicY, m_numDigY, ymi, yma, fy,
            0.2 * m_width / m_height, m_logY);
}

void FrameItem::autoScale(bool autoNorm)
{
  double fx, fy;
  if (calcAutoScale(autoNorm, childItems(), fx, fy)) {
    update();
    foreach (QGraphicsItem* item, childItems())
      switch (item->type()) {
        case ArrayItem::Type: {
          ArrayItem* it = qgraphicsitem_cast<ArrayItem*>(item);
          if (!it->isNormalized(fx, fy)) {
            it->normalize(fx, fy);
            it->update();
          }
          break;
        }
        case FunItem::Type: {
          FunItem* it = qgraphicsitem_cast<FunItem*>(item);
          if (!it->isNormalized(fx, fy)) {
            it->normalize(fx, fy);
            it->update();
          }
          break;
        }
        case ParFunItem::Type: {
          ParFunItem* it = qgraphicsitem_cast<ParFunItem*>(item);
          if (!it->isNormalized(fx, fy)) {
            it->normalize(fx, fy);
            it->update();
          }
          break;
        }
        case SplineItem::Type: {
          SplineItem* it = qgraphicsitem_cast<SplineItem*>(item);
          if (!it->isNormalized(fx, fy)) {
            it->normalize(fx, fy);
            it->update();
          }
          break;
        }
      }
  }
}

QRectF FrameItem::boundingRect() const
{
  paintCond(0);
  if (bRect.isNull())
    bRect = QGraphicsRectItem::boundingRect();
  else
    bRect |= QGraphicsRectItem::boundingRect();
  return bRect;
}

bool FrameItem::calcAutoScale(bool autoNorm,
                              const QList<QGraphicsItem*>& items, double& fx,
                              double& fy)
{
  int iexmin = 0;
  int ieymin = 0;
  if (autoNorm) {
    double fxt, fyt;
    int iext, ieyt;
    iexmin = ieymin = 999;
    foreach (QGraphicsItem* itm, items) {
      if (itm->isVisible()) {
        iext = iexmin;
        ieyt = ieymin;
        exponent(itm, iext, ieyt, fxt, fyt, m_logX);
        if (iext < iexmin) {
          iexmin = iext;
          fx = fxt;
        }
        if (ieyt < ieymin) {
          ieymin = ieyt;
          fy = fyt;
        }
      }
    }
  } else
    fx = fy = 1;
  bool success = (iexmin != 999) && (ieymin != 999);
  if (success) {
    setExpX(iexmin);
    setExpY(ieymin);
    double xmin = 1e300;
    double ymin = xmin;
    double xmax = -xmin;
    double ymax = xmax;
    foreach (QGraphicsItem* itm, items) {
      if (itm->isVisible()) {
        double xmint = xmin;
        double ymint = ymin;
        double xmaxt = xmax;
        double ymaxt = ymax;
        limits(itm, xmint, xmaxt, ymint, ymaxt, m_logX);
        xmin = qMin(xmin, xmint);
        xmax = qMax(xmax, xmaxt);
        ymin = qMin(ymin, ymint);
        ymax = qMax(ymax, ymaxt);
      }
    }
    autoScale(xmin, xmax, fx, ymin, ymax, fy);
  }
  return success;
}

QPointF FrameItem::mapFromUser(const QPointF& p) const
{
  return md.map(QPointF(m_logX ? log10(p.x()) : p.x(), m_logY ? log10(p.y()) :
                                                                p.y()));
}

QPointF FrameItem::mapToUser(const QPointF& pos) const
{
  QPointF p = m_inv.map(pos);
  if (m_logX)
    p.setX(pow(10, p.x()));
  if (m_logY)
    p.setY(pow(10, p.y()));
  return p;
}

QRectF FrameItem::mapToUser(const QRectF& rect) const
{
  return QRectF(mapToUser(rect.topLeft()), mapToUser(rect.bottomRight()));
}

void FrameItem::paint(QPainter* painter,
                      const QStyleOptionGraphicsItem* option, QWidget* widget)
{
  paintCond(painter);
  QGraphicsRectItem::paint(painter, option, widget);
}

QPointF FrameItem::rectPos() const
{
  return QPointF(m_x0, m_y0);
}

void FrameItem::setProperties(const QTransform& m)
{
  m_matrix = m;
  setProperties();
}

QPointF FrameItem::steps() const
{
  double x = 1 / (md.m11() * m_normX);
  double y = -1 / (md.m22() * m_normY);
  return QPointF(m_logX ? pow(10, x) : x, m_logY ? pow(10, y) : y);
}

void FrameItem::update()
{
  setProperties();
  QGraphicsRectItem::update();
}

void FrameItem::updatePos(QPointF p)
{
  m_x0 = p.x();
  m_y0 = p.y();
}

bool FrameItem::valid(double x, double y) const
{
  return (valid(x, m_logX) && valid(y, m_logY));
}

void FrameItem::drawText(QPainter* painter, const QString& text, double x,
                         double y, TextItem::Alignment alignment,
                         double dir) const
{
  if (!text.isEmpty()) {
    QTextDocument doc;
    TextItem::setText(doc, text, m_colorFrame);
    QTextFrameFormat tf = doc.rootFrame()->frameFormat();
    tf.setMargin(0);
    doc.rootFrame()->setFrameFormat(tf);
    QFont f("Helvetica");
    int w = qRound(0.033333 * rect().width() * m_relSize);
    if (w)
      f.setPixelSize(w);
    doc.setDefaultFont(f);
    QRectF r = doc.documentLayout()->blockBoundingRect(doc.begin());
    r.moveTopLeft(r.topLeft() + QPointF(-0.5 * (alignment - 1) * r.width(),
                                        -0.65 * QFontMetricsF(f).ascent()));
    QPointF p = md.map(QPointF(x, y));
    if (painter) {
      painter->save();
      painter->translate(p);
      if (dir)
        painter->rotate(-dir);
      painter->translate(r.topLeft());
      doc.drawContents(painter);
      painter->restore();
    } else {
      if (dir)
        r = QTransform().rotate(-dir).mapRect(r);
      r.translate(p);
      if (bRect.isNull())
        bRect = r;
      else
        bRect |= r;
    }
  }
}

double FrameItem::incLog(double& x, double f) const
{
  double mant = x - int(x + 1e-6 * fabs(x));
  if (fabs(mant) < 1e-9)
    mant = 0;
  if (mant < 0)
    ++mant;
  if (mant < 0.1)
    x += log10(f);
  else
    if (mant < 0.4)
      x += log10(2.5);
    else
      if (mant < 0.6)
        x += log10(10.0 / 3);
      else
        x += log10(2);
  return mant;
}

void FrameItem::mouseReleaseEvent(QGraphicsSceneMouseEvent* event)
{
  QGraphicsRectItem::mouseReleaseEvent(event);
  ((KplScene*)scene())->updatePos(this);
}

int FrameItem::number(double v, int numDig, QString& label) const
{
  label = QString("%1").arg(v, 0, 'f', qMax(0, numDig));
  if (numDig == 0)
    label += ".";
  return label.length();
}

void FrameItem::paintCond(QPainter* painter) const
{
  if (painter)
    painter->setPen(pen());
  else
    bRect = QRectF();
  double deltaX = xmaxl - xminl;
  double deltaY = ymaxl - yminl;
  double dxr = m_relSize * deltaX;
  double dydx = md.m11() / md.m22();
  double dyr = dxr * dydx;
  double xtc = 0;
  double ytc = 0;
  if (m_gridMode) {
    xtc = tick(m_xTic, m_logX);
    ytc = tick(m_yTic, m_logY);
    if (painter) {
      QPen p = pen();
      p.setColor(m_colorGrid);
      painter->setPen(changePenWidth(p, p.widthF() * penWidthFactor(GridPen) /
                      penWidthFactor(FramePen)));
      double dx = 0.99999 * deltaX;
      double dy = 0.99999 * deltaY;
      if (m_gridMode <= AxesWithLabels) {
        dx = 0.02 * deltaX;
        dy = -dx * dydx;
      }
      double dx2 = 0.5 * dx;
      double dy2 = 0.5 * dy;
      if (m_xOffset && (!m_logX) && m_gridMode <= AxesWithLabels &&
          m_minTicX > 1) {
        double d = xtc / m_minTicX;
        for (double xm = xminl + m_xOffset - d; xm > xminl; xm -= d)
          if (xm < xmaxl) {
            if (m_xBottom)
              painter->drawLine(md.map(QLineF(xm, yminl, xm, yminl + dy2)));
            if (m_xTop)
              painter->drawLine(md.map(QLineF(xm, ymaxl, xm, ymaxl - dy2)));
          }
      }
      for (double x = xminl + m_xOffset; x < xmaxl;)
        if (m_logX) {
          double xa = x;
          double mant = incLog(x, xtc);
          if (x < xmaxl) {
            if (m_xBottom)
              painter->drawLine(md.map(QLineF(x, yminl, x, yminl + dy)));
            if (m_xTop)
              painter->drawLine(md.map(QLineF(x, ymaxl, x, ymaxl - dy)));
          }
          if (m_minTicX > 1) {
            double xd = pow(10, xa - mant);
            for (;;) {
              xa = log10(pow(10, xa) + xd);
              if ((xa >= x) || (xa >= xmaxl))
                break;
              if (m_xBottom)
                painter->drawLine(md.map(QLineF(xa, yminl, xa,
                                  yminl + (m_gridMode <= AxesWithLabels ?
                                           dy2 : dy))));
              if (m_xTop)
                painter->drawLine(md.map(QLineF(xa, ymaxl, xa,
                                  ymaxl - (m_gridMode <= AxesWithLabels ?
                                           dy2 : dy))));
            }
          }
        } else {
          if (m_gridMode <= AxesWithLabels && m_minTicX > 1) {
            double xm = x;
            for (int i = 1; i < m_minTicX; i++)
              if ((xm += xtc / m_minTicX) < xmaxl) {
                if (m_xBottom)
                  painter->drawLine(md.map(QLineF(xm, yminl, xm,
                                                  yminl + dy2)));
                if (m_xTop)
                  painter->drawLine(md.map(QLineF(xm, ymaxl, xm,
                                                  ymaxl - dy2)));
              }
          }
          if (x > xminl) {
            if (m_xBottom)
              painter->drawLine(md.map(QLineF(x, yminl, x, yminl + dy)));
            if (m_xTop)
              painter->drawLine(md.map(QLineF(x, ymaxl, x, ymaxl - dy)));
          }
          x += xtc;
        }
      if (m_yOffset && (!m_logY) && m_gridMode <= AxesWithLabels &&
          m_minTicY > 1) {
        double d = ytc / m_minTicY;
        for (double ym = yminl + m_yOffset - d; ym > yminl; ym -= d)
          if (ym < ymaxl) {
            if (m_yLeft)
              painter->drawLine(md.map(QLineF(xminl, ym, xminl + dx2, ym)));
            if (m_yRight)
              painter->drawLine(md.map(QLineF(xmaxl, ym, xmaxl - dx2, ym)));
        }
      }
      for (double y = yminl + m_yOffset; y < ymaxl;)
        if (m_logY) {
          double ya = y;
          double mant = incLog(y, ytc);
          if (y < ymaxl) {
            if (m_yLeft)
              painter->drawLine(md.map(QLineF(xminl, y, xminl + dx, y)));
            if (m_yRight)
              painter->drawLine(md.map(QLineF(xmaxl, y, xmaxl - dx, y)));
          }
          if (m_minTicY > 1) {
            double yd = pow(10, ya - mant);
            for (;;) {
              ya = log10(pow(10, ya) + yd);
              if ((ya >= y) || (ya >= ymaxl))
                break;
              if (m_yLeft)
                painter->drawLine(md.map(QLineF(xminl, ya, xminl +
                                                (m_gridMode <= AxesWithLabels ?
                                                 dx2 : dx),
                                                ya)));
              if (m_yRight)
                painter->drawLine(md.map(QLineF(xmaxl, ya, xmaxl -
                                                (m_gridMode <= AxesWithLabels ?
                                                 dx2 : dx),
                                                ya)));
            }
          }
        } else {
          if (m_gridMode <= AxesWithLabels && m_minTicY > 1) {
            double ym = y;
            for (int i = 1; i < m_minTicY; i++)
              if ((ym += ytc / m_minTicY) < ymaxl) {
                if (m_yLeft)
                  painter->drawLine(md.map(QLineF(xminl, ym, xminl + dx2,
                                                  ym)));
                if (m_yRight)
                  painter->drawLine(md.map(QLineF(xmaxl, ym, xmaxl - dx2,
                                                  ym)));
              }
          }
          if (y > yminl) {
            if (m_yLeft)
              painter->drawLine(md.map(QLineF(xminl, y, xminl + dx, y)));
            if (m_yRight)
              painter->drawLine(md.map(QLineF(xmaxl, y, xmaxl - dx, y)));
          }
          y += ytc;
        }
    }
  }
  if ((m_gridMode == AxesWithLabels) || (m_gridMode == GridWithLabels)) {
    if (painter)
      painter->setPen(pen());
    double d = 0.03225 * dyr;
    double yB = 0;
    double yT = 0;
    if (m_xBottom)
      yB = yminl + d;
    if (m_xTop)
      yT = ymaxl - d;
    bool e = (m_logX && (xtc > 9.9) &&
             ((xmaxl > 2.999999) || (xminl < -2.999999)));
    int nd = m_numDigX;
    double xe = xmaxl + 1e-3 * deltaX;
    double tol = 1e-12 * xtc;
    QString s;
    for (double x = xminl + m_xOffset; x <= xe;) {
      if (x > xmaxl)
         x = xmaxl;
      if (e) {
        s = "10<sup>" + QString::number(qRound(x)) + "</sup> ";
        if (m_xBottom)
          drawText(painter, s, x, yB, TextItem::AlignCentered);
        if (m_xTop)
          drawText(painter, s, x, yT, TextItem::AlignCentered);
      } else {
        if (m_logX)
          nd = x < 0 ? (1 - int(x + 0.01)) : -1;
        number(m_logX ? pow(10, x) : (fabs(x) < tol ? 0 : x), nd, s);
        if (m_xBottom)
          drawText(painter, s, x, yB, TextItem::AlignCentered);
        if (m_xTop)
          drawText(painter, s, x, yT, TextItem::AlignCentered);
      }
      if (m_logX)
        incLog(x, xtc);
      else
        x += xtc;
    }
    d = 0.025 * dxr;
    double xL = 0;
    double xR = 0;
    if (m_yLeft)
      xL = xminl - d;
    if (m_yRight)
      xR = xmaxl + d;
    e = (m_logY && (ytc > 9.9) && ((ymaxl > 2.999999) || (yminl < -2.999999)));
    nd = m_numDigY;
    double ye = ymaxl + 1e-3 * deltaY;
    tol = 1e-12 * ytc;
    for (double y = yminl + m_yOffset; y <= ye;) {
      if (y > ymaxl)
        y = ymaxl;
      if (e) {
        s = "10<sup>" + QString::number(qRound(y)) + "</sup> ";
        if (m_yLeft)
          drawText(painter, s, xL, y, TextItem::AlignRight);
        if (m_yRight)
          drawText(painter, s, xR, y, TextItem::AlignLeft);
      } else {
        if (m_logY)
          nd = y < 0 ? (1 - int(y + 0.01)) : -1;
        number(m_logY ? pow(10, y) : (fabs(y) < tol ? 0 : y), nd, s);
        if (m_yLeft)
          drawText(painter, s, xL, y, TextItem::AlignRight);
        if (m_yRight)
          drawText(painter, s, xR, y, TextItem::AlignLeft);
      }
      if (m_logY)
        incLog(y, ytc);
      else
        y += m_yTic;
    }
  }
  if (painter)
    painter->setPen(pen());
  double dx = 0.025 * dxr;
  double dy = 0.09475 * dyr;
  double x = 0.5 * (xminl + xmaxl);
  double y = 0;
  QString s;
  if (m_xBottom || m_xTop) {
    if (m_textX.contains("</") || (!m_textX.contains('#')))
      s = m_expX ?
          ("10<sup>" + QString::number(m_expX) + "</sup> " + m_textX) :
          m_textX;
    else
      s = m_expX ? ("10#NU" + QString::number(m_expX) + " #NN" + m_textX) :
                    m_textX;
  }
  if (m_xBottom)
    drawText(painter, s, x, yminl + dy, TextItem::AlignCentered);
  if (m_xTop)
    drawText(painter, s, x, ymaxl - dy, TextItem::AlignCentered);
  int l = 3;
  if (m_yLeft || m_yRight) {
    if (m_logY)
      if ((ymaxl > 2.999999) || (yminl < -2.999999))
        l += (yminl < 0) ? 4 : 3;
      else
        l += 1 + int(qMax(ymaxl, 1.99 - yminl));
    else
      l += qMax(number(yminl, m_numDigY, s), number(ymaxl, m_numDigY, s));
    if (m_textY.contains("</") || (!m_textY.contains('#')))
      s = m_expY ?
          ("10<sup>" + QString::number(m_expY) + "</sup> " + m_textY) :
          m_textY;
    else
      s = m_expY ? ("10#NU" + QString::number(m_expY) + " #NN" + m_textY) :
                    m_textY;
    y = 0.5 * (yminl + ymaxl);
  }
  if (m_yLeft)
    drawText(painter, s, xminl - (0.75 * l + 0.71) * dx, y,
             TextItem::AlignCentered, 90);
  if (m_yRight)
    drawText(painter, s, xmaxl + (0.75 * l + 0.71) * dx, y,
             TextItem::AlignCentered, -90);
  drawText(painter, m_header, x, ymaxl - (m_xTop ? 0.1875 : 0.04275) * dyr,
           TextItem::AlignCentered);
}

void FrameItem::setProperties()
{
  xminl = m_logX ? log10(qMax(m_xMin, 1e-300)) : m_xMin;
  xmaxl = m_logX ? log10(qMax(m_xMax, 1.1e-300)) : m_xMax;
  yminl = m_logY ? log10(qMax(m_yMin, 1e-300)) : m_yMin;
  ymaxl = m_logY ? log10(qMax(m_yMax, 1.1e-300)) : m_yMax;
  QRectF r = m_matrix.mapRect(QRectF(m_x0, m_y0, m_width, m_height));
  setRect(r);
  QPen pen = createPen(this, m_colorFrame, FramePen);
  double w = r.width();
  double d = w * m_relSize * penWidthFactor(FramePen);
  setPen(changePenWidth(pen, d));
  double h = r.height();
  double dy = h + r.y();
  md = QTransform(w / (xmaxl - xminl), 0, 0, h / (yminl - ymaxl), r.x(), dy);
  md.translate(-xminl, -yminl);
  m_inv = QTransform().translate(xminl, yminl).scale(1 / md.m11(),
                                 1 / md.m22()).translate(-r.x(), -dy);
  d *= 0.5;
  cRect = r.adjusted(d, d, -d, -d);
}

double FrameItem::tick(double tic, bool log) const
{
  if (log) {
    if (tic < 2.5)
      tic = 2;
    else
      if (tic < 5.5)
        tic = 3;
      else
        if (tic < 99)
          tic = 10;
  }
  return tic;
}

void FrameItem::writePlo(KConfigGroup* plo) const
{
  plo->writeEntry("Type", "FRAMEITEM");
  KplItem::writePlo(plo, isVisible());
  if (m_logX)
    plo->writeEntry("logx", m_logX);
  if (m_logY)
    plo->writeEntry("logy", m_logY);
  if (!m_xBottom)
    plo->writeEntry("xBottom", m_xBottom);
  if (m_xTop)
    plo->writeEntry("xTop", m_xTop);
  if (!m_yLeft)
    plo->writeEntry("yLeft", m_yLeft);
  if (m_yRight)
    plo->writeEntry("yRight", m_yRight);
  if (m_collapsed)
    plo->writeEntry("collapsed", m_collapsed);
  plo->writeEntry("ndigx", m_numDigX);
  plo->writeEntry("ndigy", m_numDigY);
  plo->writeEntry("mticx", m_minTicX);
  plo->writeEntry("mticy", m_minTicY);
  plo->writeEntry("gridmode", int(m_gridMode));
  plo->writeEntry("iex", m_expX);
  plo->writeEntry("iey", m_expY);
  plo->writeEntry("colframe", QString::number(rgbQt1(m_colorFrame), 16));
  plo->writeEntry("colgrid", QString::number(rgbQt1(m_colorGrid), 16));
  plo->writeEntry("xw0", m_x0);
  plo->writeEntry("xw", m_width);
  plo->writeEntry("yw0", m_y0);
  plo->writeEntry("yw", m_height);
  plo->writeEntry("xmi", m_xMin);
  plo->writeEntry("xma", m_xMax);
  plo->writeEntry("ymi", m_yMin);
  plo->writeEntry("yma", m_yMax);
  plo->writeEntry("xtic", m_xTic);
  plo->writeEntry("ytic", m_yTic);
  if (m_xOffset)
    plo->writeEntry("xOff", m_xOffset);
  if (m_yOffset)
    plo->writeEntry("yOff", m_yOffset);
  plo->writeEntry("sx", m_textX);
  plo->writeEntry("sy", m_textY);
  plo->writeEntry("sh", m_header);
}
