/***************************************************************************
                          funitem.cpp  -  description
                             -------------------

    This file is a part of kpl - a program for graphical presentation of
    data sets and functions.

    begin                : Sun Aug 29 1999
    copyright            : (C) 2015 by Werner Stille
    email                : stille@uni-freiburg.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef NAN
#include <stdlib.h>
#endif
#include <QFileInfo>
#if QT_VERSION >= 0x040700
#include <QElapsedTimer>
#endif
#include <QLibrary>
#include <QPainter>
#include <QProcess>
#include <QTextStream>
#include <KConfigGroup>
#include <KLocalizedString>
#include <KMessageBox>
#include <ThreadWeaver/JobCollection>
#include <ThreadWeaver/Weaver>
#include "arrayitem.h"
#include "calcjob.h"
#include "frameitem.h"
#include "funitem.h"
#include "kplscene.h"
#include "lmfit.h"

FunItem::FunItem() :
 funcY(0), logXo(false), m_tMin(0), m_tMax(0), m_deltaT(0), tMino(0), tMaxo(0),
 deltaTo(0), funcYo(0), m_libY(0)
{
  init();
}

FunItem::FunItem(const FunItem& a) :
 ScaledItem(a), funcY(a.funcY), logXo(a.logXo), m_tMin(a.m_tMin),
 m_tMax(a.m_tMax), m_deltaT(a.m_deltaT), tMino(a.tMino), tMaxo(a.tMaxo),
 deltaTo(a.deltaTo), funcYo(a.funcYo), m_tValues(a.m_tValues),
 m_yValues(a.m_yValues), m_nameY(a.m_nameY), m_urlY(a.m_urlY), m_libY(0)
{
  memcpy(m_parY, a.m_parY, sizeof(m_parY));
  memcpy(parYo, a.parYo, sizeof(parYo));
  if (a.m_libY)
    getFuncAddr(m_urlY.path(), m_nameY, &m_libY, &funcY);
}

FunItem::FunItem(const KConfigGroup* plo, const KUrl& urlPlo, unsigned color,
                 Symbol symbol, QGraphicsItem* parent) :
 ScaledItem(plo, symbol, color, parent), logXo(false),
 tMino(0), tMaxo(0), deltaTo(0), funcYo(0)
{
  m_tMin = plo->readEntry("xfmin", 0.0);
  m_tMax = plo->readEntry("xfmax", m_tMin + 1);
  m_deltaT = plo->readEntry("dx", 0.0);
  memset(parYo, 0, sizeof(parYo));
  stringToPar(plo->readEntry("p"), m_parY);
  m_nameY = plo->readEntry("name", "");
  m_urlY = absoluteUrl(plo->readEntry("path", ""), urlPlo);
  getFuncAddr(m_urlY.path(), m_nameY, &m_libY, &funcY);
}

FunItem::FunItem(const KUrl& path, const QString& name, unsigned color,
                 Symbol symbol, double fx, double fy, QGraphicsItem* parent,
                 double xMin, double xMax, double deltaX, bool visible,
                 FillStyle fillStyle, double xOffset, double yOffset,
                 double relSize) :
 ScaledItem(fillStyle, symbol, color, fx, fy, xOffset, yOffset, relSize,
            parent), funcY(0), logXo(false), m_tMin(xMin), m_tMax(xMax),
 m_deltaT(deltaX), tMino(0), tMaxo(0), deltaTo(0), funcYo(0), m_nameY(name),
 m_urlY(path), m_libY(0)
{
  if (!visible)
    hide();
  init();
  if (!m_urlY.url().isEmpty())
    getFuncAddr(m_urlY.path(), m_nameY, &m_libY, &funcY);
}

FunItem::~FunItem()
{
  unload();
}

QRectF FunItem::boundingRect() const
{
  if (isVisible() && bRect.isNull())
    paintCond(0);
  return bRect;
}

void FunItem::calcTable(const double* x, double* values,
                        double (*func)(double, const double*),
                        const double* par, int n, QObject* parent)
{
  int nJobs =
    qMin(n, ThreadWeaver::Weaver::instance()->maximumNumberOfThreads());
#if QT_VERSION >= 0x040700
  QElapsedTimer timer;
  timer.start();
#endif
  if (nJobs > 1) {
    JobCollection jobs;
    int i0 = 0;
    CalcJob* jobList[nJobs];
    for (int i = 0; i < nJobs; ++i) {
      int nt = (n - i0) / (nJobs - i);
      jobs.addJob(jobList[i] = new CalcJob(&x[i0], &values[i0], func, par, nt,
                                           parent));
      i0 += nt;
    }
    ThreadWeaver::Weaver::instance()->enqueue(&jobs);
    ThreadWeaver::Weaver::instance()->finish();
    for (int i = 0; i < nJobs; ++i)
      delete jobList[i];
  } else
    for (int i = 0; i < n; i++)
      values[i] = FrameItem::valid(x[i]) ? ((*func)(x[i], par)) :
#ifdef NAN
        NAN;
#else
        atof("nan");
#endif
#if QT_VERSION >= 0x040700
  long tns = timer.nsecsElapsed();
  qDebug("FunItem:calcTable(): %d points,  %d threads, %4.4g s, %4.4g points / s",
         n, nJobs, 1e-9 * tns, 1e9 * n / tns);
#endif
}

int FunItem::calcTable() const
{
  return calcTable(logXo);
}

int FunItem::calcTable(bool logX) const
{
  if (!funcY)
    return 0;
  bool newv = m_tValues.isEmpty();
  double d;
  int n;
  if (m_deltaT) {
    n = int(fabs((m_tMax - m_tMin) / m_deltaT)) + 1;
    d = logX ? (log10(m_tMax / m_tMin) / (n - 1)) : m_deltaT;
  } else {
    n = 101;
    d = 0.01 * (logX ? log10(m_tMax / m_tMin) : (m_tMax - m_tMin));
  }
  if (newv || (m_tMin != tMino) || (m_tMax != tMaxo) || (d != deltaTo) ||
      (logX != logXo)) {
    m_tValues.resize(n);
    int n1 = n - 1;
    for (int i = 0; i < n; i++)
      m_tValues[i] = (i == n1) ? m_tMax :
                                 (logX ? pow(10.0, log10(m_tMin) + i * d) :
                                         (m_tMin + i * d));
  }
  bool pchanged = parChanged(m_parY, parYo);
  if (pchanged || newv || (m_tMin != tMino) || (m_tMax != tMaxo) ||
      (d != deltaTo) || (funcY != funcYo) || (logX != logXo))
    m_yValues.resize(n);
  if (pchanged || newv || (m_tMin != tMino) || (m_tMax > tMaxo) ||
      (d != deltaTo) || (funcY != funcYo) || (logX != logXo))
    calcTable(m_tValues.data(), m_yValues.data(), funcY, m_parY, n);
  if (pchanged)
    memcpy(parYo, m_parY, sizeof(parYo));
  funcYo = funcY;
  tMino = m_tMin;
  tMaxo = m_tMax;
  deltaTo = d;
  logXo = logX;
  return n;
}

bool FunItem::errorStructChanged(const DataErrorStruct& e1,
                                 const DataErrorStruct& e2)
{
  if ((e1.fitErrCol != e2.fitErrCol) || (e1.errModArg != e2.errModArg) ||
      (e1.errModName != e2.errModName) || (e1.errModPath != e2.errModPath))
    return true;
  return parChanged(e1.parErrMod, e2.parErrMod);
}

void FunItem::exponent(int& expX, int& expY, double& fx, double& fy,
                       bool logX) const
{
  if (funcY) {
    ScaledItem::exponent(qMax(fabs(m_tMin), fabs(m_tMax)), expX, fx);
    double xmin, xmax, ymin, ymax;
    limits(xmin, xmax, ymin, ymax, logX);
    ScaledItem::exponent(qMax(fabs(ymin), fabs(ymax)), expY, fy);
  }
}

void FunItem::exportTable(QTextStream& stream, char separator, char format,
                          int precision) const
{
  int n = calcTable(logXo);
  for (int i = 0; i < n; i++)
    stream << QString::number(m_normX * m_tValues[i], format, precision)
           << separator
           << QString::number(m_normY * m_yValues[i], format, precision)
           << "\n";
}

int FunItem::fit(const QList<const ArrayItem*>* ai, QList<FunItem*>* fi,
                 const bool* bFit, double* par, QVector<double>* resVector,
                 QVector<double>* sigma, FunItem::DataErrorStruct* err,
                 bool nonLin, double tol, int itMax, LMFit* lma,
                 double* chiSquare, double* corr, double* pErr, double* avgErr,
                 QObject* parent)
{
  foreach (FunItem* f, *fi)
    if (!f->funcY)
      return -101;
  int i, k;
  int np = 0;
  for (i = 0; i < NumParMax; i++)
    if (bFit[i])
      np++;
  if (!np)
    return -102;
  int n = 0;
  foreach (const ArrayItem* a, *ai)
    n += a->count();
  QVector<double> *fvec, *sig;
  if (resVector) {
    fvec = resVector;
    fvec->resize(n);
  } else
    fvec = new QVector<double>(n);
  if (sigma) {
    sig = sigma;
    sig->resize(n);
  } else
    sig = new QVector<double>(n);
  int j = 0;
  for (k = 0; k < (int) ai->count(); k++) {
    const ArrayItem* a = ai->at(k);
    if ((!err) || err[k].fitErrCol) {
      for (i = 0; i < a->count(); i++)
        (*sig)[j++] = a->data()[a->colErr()][a->rowStart() + i];
    } else {
      double (*fktErrMod)(double, const double*);
      QLibrary* lib;
      if (!getFuncAddr(err[k].errModPath, err[k].errModName, &lib, &fktErrMod))
        return -103;
      for (i = 0; i < a->count(); i++)
        (*sig)[j++] = fktErrMod(a->data()[err[k].errModArg ?
                                          a->colY() : a->colX()]
                                [a->rowStart() + i], err[k].parErrMod);
      lib->unload();
      delete lib;
    }
  }
  LMFit* lm = lma ? lma : new LMFit(ai, fi, bFit, par, parent);
  lm->setSigma(sig->data());
  QVector<double> guess(np);
  j = 0;
  for (i = 0; i < NumParMax; i++)
    if (bFit[i])
      guess[j++] = par[i];
  bool aborted = false;
  lm->setUserBreak(aborted);
  QVector<int> ip(NumParMax);
  ip[0] = -1;
  int info = 0;
  QVector<double> fjac(n * np);
  if (nonLin) {
    lm->lmdif(n, np, guess.data(), fvec->data(), tol, itMax, &info,
              fjac.data(), ip.data());
    aborted = (ip[0] < 0);
    if (info >= 8)
      info = 4;
  } else
    aborted = lm->linFit(n, np, guess.data(), fvec->data(), fjac.data(),
                         ip.data());
  if (!aborted) {
    j = 0;
    for (i = 0; i < NumParMax; i++)
      if (bFit[i])
        par[i] = guess[j++];
    if (chiSquare) {
      double chi = LMFit::enorm(n, fvec->data());
      *chiSquare = chi * chi;
    }
    if (corr) {
      QVector<double> alpha((np + n - 1) * np);
      for (i = 0; i < np; i++) {
        for (j = 0; j < np; j++) {
          alpha[ip[i] + ip[j] * np] = 0;
          for (int k = 0; k <= j; k++)
            if (k <= i)
              alpha[ip[i] + ip[j] * np] += fjac[i * n + k] * fjac[j * n + k];
        }
      }
      LMFit::minv(alpha.data(), corr, np);
      int ij = 0;
      QVector<double> diag(np);
      for (i = 0; i < np; i++) {
        diag[i] = sqrt(corr[ij]);
        ij += np + 1;
      }
      k = 0;
      for (i = 0; i < np; i++)
        for (j = 0; j < np; j++) {
          if (double temp = diag[i] * diag[j])
            corr[k] /= temp;
          k++;
        }
      int ny = n - np;
      if (chiSquare && pErr && ny) {
        double temp = sqrt(*chiSquare / ny);
        for (i = 0; i < np; i++)
          diag[i] *= temp;
        k = 0;
        for (i = 0; i < NumParMax; i++)
          pErr[i] = bFit[i] ? diag[k++] : 0;
      }
    }
    if (avgErr) {
      *avgErr = 0;
      for (i = 0; i < n; i++)
        if (FrameItem::valid((*sig)[i])) {
          double temp = (*fvec)[i] * (*sig)[i];
          *avgErr += temp * temp;
        }
      *avgErr = sqrt(*avgErr / n);
    }
  }
  if (!lma)
    delete lm;
  if (!sigma)
    delete sig;
  if (!resVector)
    delete fvec;
  if (aborted)
    info = -100;
  return info;
}

QStringList FunItem::funcNames(const KUrl& url)
{
  QStringList list;
  QRegExp rx("^[0-9a-f]{8,} T ([^_\\s]+\\S*)");
  QFileInfo fInf(url.path());
  QString path = fInf.absolutePath() + "/" + fInf.baseName();
  QFile f(path + ".def");
  if (f.open(QIODevice::ReadOnly)) {
    while (!f.atEnd())
      if (rx.indexIn(f.readLine()) != -1)
        list << rx.capturedTexts()[1];
    f.close();
  } else {
    QProcess p;
    p.start(QString("nm \"%1\"").arg(url.path()));
    p.waitForFinished();
    while (!p.atEnd())
      if (rx.indexIn(p.readLine()) != -1)
        list << rx.capturedTexts()[1];
    p.close();
  }
  if (list.isEmpty()) {
    f.setFileName(path + ".c");
    bool open = f.open(QIODevice::ReadOnly);
    if (!open) {
      f.setFileName(path + ".cpp");
      open = f.open(QIODevice::ReadOnly);
    }
    if (open) {
      QRegExp rx2("^\\s?double\\s+([^_\\s]+\\S*)\\s?\\(\\s?double\\s?");
      while (!f.atEnd())
        if (rx2.indexIn(f.readLine()) != -1)
          list << rx2.capturedTexts()[1];
      f.close();
      list.sort();
    }
  }
  return list;
}

bool FunItem::getFuncAddr(const QString& path, const QString& name,
                          QLibrary** lib,
                          double (**func)(double, const double*))
{
  *lib = 0;
  *func = 0;
  if (path.isEmpty() || name.isEmpty())
    KMessageBox::sorry(0, i18n("no library or no function specified!"));
  else {
    *lib = new QLibrary(path);
    if (!(*lib)->load())
      KMessageBox::error(0, path + "\n" + i18n("library cannot be opened!"));
    else
      if ((*func = (double (*)(double, const double *))
           (*lib)->resolve(name.toLatin1())))
        return true;
      else
        KMessageBox::error(0, path + ": " + name + "\n" +
                           i18n("function name cannot be resolved!"));
    delete *lib;
    *lib = 0;
  }
  return false;
}

void FunItem::limits(double& xMin, double& xMax, double& yMin,
                     double& yMax, bool logX) const
{
  if (funcY) {
    xMin = m_tMin + m_xOffset;
    xMax = m_tMax + m_xOffset;
    int n = calcTable(logX);
    ScaledItem::limits(yMin, yMax, m_yValues.data(), n);
    yMin += m_yOffset;
    yMax += m_yOffset;
  }
}

void FunItem::paint(QPainter* painter, const QStyleOptionGraphicsItem*,
                    QWidget*)
{
  paintCond(painter);
}

bool FunItem::parChanged(const double* p1, const double* p2)
{
  for (int i = 0; i < NumParMax; ++i)
    if (p1[i] != p2[i])
      return true;
  return false;
}

QString FunItem::parToString(const double* par, char frm, int precision)
{
  QStringList list;
  for (int i = 0; i < NumParMax; ++i)
    list.append(QString::number(par[i], frm, precision));
  return list.join(" ");
}

void FunItem::setDashOffset()
{
  if (!m_fillStyle  && (m_symbol > 1)) {
    m_pen.setDashOffset(m_dashOffset);
    double s = 0;
    foreach (double d, m_pen.dashPattern())
      s += d;
    FrameItem* frame = qgraphicsitem_cast<FrameItem*>(parentItem());
    m_dashOffset += length(m_tValues.data(), m_yValues.data(),
                           calcTable((frame && frame->isVisible()) ?
                                     frame->logX() : false)) / m_pen.widthF();
    m_dashOffset -= int(m_dashOffset / s) * s;
  }
}

bool FunItem::setFuncY(const KUrl& url, const QString& name)
{
  m_urlY = url;
  m_nameY = name;
  unload();
  return getFuncAddr(url.path(), name, &m_libY, &funcY);
}

void FunItem::setPar(int i, double value, bool yFunc)
{
  if (yFunc)
    m_parY[i] = value;
}

QPainterPath FunItem::shape() const
{
  if (isVisible() && bPath.isEmpty())
    paintCond(0);
  return bPath;
}

void FunItem::stringToPar(const QString& s, double* par)
{
  QStringList list = s.split(' ');
  int cnt = list.count();
  for (int i = 0; i < NumParMax; i++)
    par[i] = (i < cnt) ? list[i].toDouble() : 0;
}

const double* FunItem::tValues() const
{
  return m_tValues.data();
};

void FunItem::writePlo(KConfigGroup* plo, const KUrl& urlPlo, bool absolute,
                       char format, int precision) const
{
  plo->writeEntry("Type", "FUNITEM");
  ScaledItem::writePlo(plo, isVisible());
  plo->writeEntry("xfmin", QString::number(m_tMin, format, precision));
  plo->writeEntry("xfmax", QString::number(m_tMax, format, precision));
  plo->writeEntry("dx", QString::number(m_deltaT, format, precision));
  plo->writeEntry("p", parToString(m_parY, format, precision));
  plo->writeEntry("name", m_nameY);
  plo->writeEntry("path", path(urlPlo, m_urlY, absolute));
}

double* FunItem::yValues() const
{
  return m_yValues.data();
};

void FunItem::init()
{
  memset(m_parY, 0, sizeof(m_parY));
  memset(parYo, 0, sizeof(parYo));
}

void FunItem::paintCond(QPainter* painter) const
{
  if (!funcY || m_symbol == -17)
    return;
  FrameItem* frame = qgraphicsitem_cast<FrameItem*>(parentItem());
  KplScene* sc = 0;
  bool logx = false;
  bool useFrame = frame && frame->isVisible();
  if (useFrame)
    logx = frame->logX();
  else
    sc = static_cast<KplScene*>(scene());
  if (painter) {
    painter->setClipRect(useFrame ? frame->clipRect() : sc->sceneRect());
    if (m_fillStyle) {
      painter->setBrush(createBrush(m_color, m_fillStyle));
      painter->setPen(Qt::NoPen);
    } else
      painter->setPen(m_pen);
  }
  plotArray(painter, m_tValues.data(), m_yValues.data(), calcTable(logx));
}

void FunItem::unload()
{
  if (m_libY) {
    m_libY->unload();
    delete m_libY;
    m_libY = 0;
  }
}
