/***************************************************************************
                          kplitem.cpp  -  description
                             -------------------

    This file is a part of kpl - a program for graphical presentation of
    data sets and functions.

    begin                : Wed Aug 25 1999
    copyright            : (C) 2015 by Werner Stille
    email                : stille@uni-freiburg.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <QColor>
#include <QDir>
#include <QFileInfo>
#include <QGraphicsRectItem>
#include <QGraphicsScene>
#include <QPainter>
#include <QPolygonF>
#include <KConfigGroup>
#include <KTemporaryFile>
#include "kplitem.h"

double KplItem::m_dashOffset = 0;

KplItem::KplItem(double relSize) : m_relSize(relSize)
{
}

KplItem::KplItem(const KplItem& a) : m_relSize(a.m_relSize)
{
}

KplItem::KplItem(const KConfigGroup* plo)
{
  readPlo(plo);
}

KplItem::~KplItem()
{
}

KUrl KplItem::absoluteUrl(const QString& path, const KUrl& urlPlo)
{
  KUrl url;
  QString s = QUrl(path).path();
  if ((!s.isEmpty()) && (s.at(0) != '/')) {
    url = urlPlo.directory(KUrl::AppendTrailingSlash |
                           KUrl::IgnoreTrailingSlash) + path;
    if (!urlPlo.isLocalFile()) {
      url.setScheme(urlPlo.scheme());
      url.setAuthority(urlPlo.authority());
    }
  } else
    url = path;
  url.cleanPath();
  return url;
}

QBrush KplItem::createBrush(unsigned color, FillStyle fillStyle)
{
  static const Qt::BrushStyle pattern[] = {Qt::NoBrush, Qt::SolidPattern,
                                           Qt::HorPattern, Qt::VerPattern,
                                           Qt::CrossPattern, Qt::BDiagPattern,
                                           Qt::FDiagPattern,
                                           Qt::DiagCrossPattern};
  return QBrush(QColor(color), pattern[fillStyle]);
}

QPen KplItem::createPen(double width, unsigned color, PenType penType,
                        Symbol symbol)
{
  QPen p(QColor(color), width, Qt::SolidLine, Qt::RoundCap, Qt::RoundJoin);
  double f = 1 / (penWidthFactor(penType) * 320);
  switch (symbol) {
    case DashLine1:
      p.setDashPattern(QVector<double>() << 8 * f << 8 * f);
      break;
    case DashLine2:
      p.setDashPattern(QVector<double>() << 2 * f << 14 * f);
      break;
    case DotLine1:
      p.setDashPattern(QVector<double>() << 0.01 * f << 15.99 * f);
      break;
    case DashLine3:
      p.setDashPattern(QVector<double>() << 12 * f << 4 * f);
      break;
    case DashDotLine:
      p.setDashPattern(QVector<double>() << 9.99 * f << 3 * f << 0.01 * f << 3 * f);
      break;
    case DashLine4:
      p.setDashPattern(QVector<double>() << 8 * f << 3 * f << 2 * f << 3 * f);
      break;
    case DashLine5:
      p.setDashPattern(QVector<double>() << 4 * f << 4 * f);
      break;
    case DotLine2:
      p.setDashPattern(QVector<double>() << 0.01 * f << 7.99 * f);
      break;
    default:
      break;
  }
  return p;
}

void KplItem::paintSymbol(QPainter* painter, const QRectF& r, Symbol symbol)
{
  switch (symbol) {
    case OpenCircle:
    case FilledCircle:
      painter->drawEllipse(r);
      break;
    case Dot:
      painter->drawEllipse(QRectF(0.2 * r.topLeft(), 0.2 * r.size()));
      break;
    case OpenDiamond:
    case FilledDiamond:
      painter->drawPolygon(QPolygonF() << QPointF(r.left(), 0)
                                       << QPointF(0, r.top())
                                       << QPointF(r.right(), 0)
                                       << QPointF(0, r.bottom()));
      break;
    case OpenSquare:
    case FilledSquare:
      painter->drawRect(r);
      break;
    case OpenTriangle:
    case FilledTriangle:
      painter->drawPolygon(QPolygonF() << QPointF(r.left(), 0.5 * r.bottom())
                                       << QPointF(r.right(), 0.5 * r.bottom())
                                       << QPointF(0, r.top()));
      break;
    case OpenHourglass:
    case FilledHourglass:
      painter->drawPolygon(QPolygonF() << r.topLeft() << r.bottomRight()
                                       << r.bottomLeft() << r.topRight());
      break;
    case OpenOctagon:
    case FilledOctagon:
      painter->drawPolygon(QPolygonF() << QPointF(r.left(), 0.5 * r.top())
                                       << QPointF(0.5 * r.left(), r.top())
                                       << QPointF(0.5 * r.right(), r.top())
                                       << QPointF(r.right(), 0.5 * r.top())
                                       << QPointF(r.right(), 0.5 * r.bottom())
                                       << QPointF(0.5 * r.right(), r.bottom())
                                       << QPointF(0.5 * r.left(), r.bottom())
                                       << QPointF(r.left(), 0.5 * r.bottom()));
      break;
    case Cross:
      painter->drawLine(r.topLeft(), r.bottomRight());
      painter->drawLine(r.bottomLeft(), r.topRight());
      break;
    case Plus:
      painter->drawLine(QPointF(r.left(), 0), QPointF(r.right(), 0));
      painter->drawLine(QPointF(0, r.top()), QPointF(0, r.bottom()));
      break;
    case Ypsilon:
      painter->drawPolyline(QPolygonF() << QPointF(r.left(), r.top())
                                        << QPointF(0, 0)
                                        << QPointF(r.right(), r.top()));
      painter->drawLine(QPointF(0, 0), QPointF(0, r.bottom()));
      break;
    case Umbrella:
      painter->drawPolyline(QPolygonF() << QPointF(0, r.bottom())
                                        << QPointF(0, r.top())
                                        << QPointF(r.left(), 0)
                                        << QPointF(r.right(), 0)
                                        << QPointF(0, r.top()));
    case Invisible:
    default:
      break;
  }
}

QPixmap KplItem::pixmap(Symbol symbol)
{
  return pixmap(createPen(2, 0, DataPen, symbol), symbol);
}

QPixmap KplItem::pixmap(QPen pen, Symbol symbol)
{
  QPixmap pm((symbol > OpenCircle) ? QSize(50, 2) : QSize(16, 16));
  pm.fill(Qt::transparent);
  QPainter painter(&pm);
  if (symbol != Invisible) {
    painter.setRenderHint(QPainter::Antialiasing);
    if ((symbol < Umbrella) || (symbol == Dot)) {
      painter.setBrush(pen.color());
      painter.setPen(Qt::NoPen);
    } else {
      pen.setWidthF(2);
      painter.setPen(pen);
    }
    if (symbol > OpenCircle)
      painter.drawLine(1, 1, 48, 1);
    else {
      painter.translate(pm.rect().center());
      paintSymbol(&painter, QRectF(-5, -5, 10, 10), symbol);
    }
  }
  return pm;
}

void KplItem::readPlo(const KConfigGroup* plo)
{
  m_relSize = plo->readEntry("relsiz", 1.0);
}

QString KplItem::path(const KUrl& urlPlo, const KUrl& url, bool absolute)
{
  bool rel = false;
  if (!absolute && !(urlPlo.isLocalFile() ^ url.isLocalFile())) {
    if (url.isLocalFile())
      rel = true;
    else {
      KUrl t(urlPlo);
      t.setPath("/");
      rel = t.isParentOf(url);
    }
  }
  return rel ? QFileInfo(urlPlo.path()).dir().relativeFilePath(url.path()) :
               (url.isLocalFile() ? url.path() : url.url());
}

unsigned KplItem::rgb(const QString& color)
{
  unsigned u = color.toUInt(0, 16);
  return QColor(int(u & 0xffL), int((u & 0xff00L) >> 8),
                int((u & 0xff0000L) >> 16)).rgb();
}

unsigned KplItem::rgbQt1(unsigned rgb)
{
  QColor col(rgb);
  return unsigned(col.red()) | (unsigned(col.green()) << 8) |
    (unsigned(col.blue()) << 16);
}

QString KplItem::tempFileName(const QString& suffix)
{
  KTemporaryFile f;
  f.setSuffix(suffix);
  f.open();
  return f.fileName();
}

void KplItem::writePlo(KConfigGroup* plo, bool visible) const
{
  if (!visible)
    plo->writeEntry("active", visible);
  plo->writeEntry("relsiz", m_relSize);
}

QPen& KplItem::changePenWidth(QPen& pen, double width) const
{
  pen.setWidthF(width);
  return pen;
}

QPen KplItem::createPen(QGraphicsItem* item, unsigned color, PenType penType,
                        Symbol symbol) const
{
  return createPen(penWidth(item, penType), color, penType, symbol);
}

double KplItem::parentWidth(const QGraphicsItem* item)
{
if (!item->scene()) qDebug("KplItem::parentWidth(): no scene!");
  return (item->parentItem() && item->parentItem()->isVisible()) ?
         static_cast<QGraphicsRectItem*>(item->parentItem())->rect().width() :
         (item->scene() ? item->scene()->width() : 100);
}

double KplItem::penWidth(const QGraphicsItem* item, PenType penType) const
{
  return parentWidth(item) * m_relSize * penWidthFactor(penType);
}

double KplItem::penWidthFactor(PenType penType)
{
  static const double f[] = {1.0 / 240, 1.0 / 480, 1.0 / 339};
  return f[penType];
}
