/***************************************************************************
                          legenditem.cpp  -  description
                             -------------------

    This file is a part of kpl - a program for graphical presentation of
    data sets and functions.

    begin                : Tue Aug 31 1999
    copyright            : (C) 2015 by Werner Stille
    email                : stille@uni-freiburg.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <QAbstractTextDocumentLayout>
#include <QGraphicsScene>
#include <QPainter>
#include <QTextFrame>
#include <KConfigGroup>
#include "arrayitem.h"
#include "frameitem.h"
#include "kplscene.h"
#include "legenditem.h"
#include "textitem.h"

LegendItem::LegendItem(double x0, double y0, unsigned colorSymbol,
                       unsigned colorText, Symbol symbol, double xOff,
                       double yOff, QGraphicsItem* parent, const QString& text,
                       bool visible, double relSize, double symbSize) :
 DecoItem(colorSymbol, x0, y0, relSize), QGraphicsItem(parent),
 m_symbol(symbol), m_textColor(colorText), m_textX(xOff), m_textY(yOff),
 m_symbolSize(symbSize), m_text(text)
{
  if (!visible)
    hide();
  setFlags(QGraphicsItem::ItemIsMovable | QGraphicsItem::ItemIsSelectable);
  init();
}

LegendItem::LegendItem(const LegendItem& a) :
 DecoItem(a), QGraphicsItem(), m_symbol(a.m_symbol),
 m_textColor(a.m_textColor), m_textX(a.m_textX), m_textY(a. m_textY),
 m_symbolSize(a.m_symbolSize), m_text(a.m_text)
{
  setZValue(a.zValue());
  setFlags(QGraphicsItem::ItemIsMovable | QGraphicsItem::ItemIsSelectable);
  setSelected(a.isSelected());
  setVisible(a.isVisible());
}

LegendItem::LegendItem(const KConfigGroup* plo, unsigned colorSymbol,
                       unsigned colorText, Symbol symbol, double frameWidth,
                       QGraphicsItem* parent) :
 DecoItem(plo, colorSymbol), QGraphicsItem(parent)
{
  if (!plo->readEntry("active", true))
    hide();
  m_symbol = Symbol(plo->readEntry("symb", int(symbol)));
  QString s1 = plo->readEntry("coltext", "");
  m_textColor = s1.isEmpty() ? colorText : rgb(s1);
  double dx = frameWidth / 15;
  m_textX = plo->readEntry("xtext", dx);
  m_textY = plo->readEntry("ytext", -0.18 * dx);
  m_symbolSize = plo->readEntry("symbsize", 1.0);
  m_text = plo->readEntry("s", "");
  setFlags(QGraphicsItem::ItemIsMovable | QGraphicsItem::ItemIsSelectable);
  init();
}

LegendItem::~LegendItem()
{
}

QRectF LegendItem::boundingRect() const
{
  double h = m_pen.widthF();
  double w;
  if (m_symbol > OpenCircle)
    w = h + 0.8 * m_textX * m_matrix.m11();
  else {
    w = h + symbRect.width();
    h += symbRect.height();
  }
  return textRect().unite(QRectF(-0.5 * w, -0.5 * h, w, h));
}

void LegendItem::paint(QPainter* painter, const QStyleOptionGraphicsItem*,
                       QWidget*)
{
  if (isSelected()) {
    painter->setPen(QPen(Qt::DashLine));
    painter->drawRect(boundingRect());
  }
  painter->setPen(m_pen);
  if (m_symbol > OpenCircle) {
    double dx = 0.4 * m_matrix.m11() * m_textX;
    painter->drawLine(QPointF(-dx, 0), QPointF(dx, 0));
  } else
    if (m_symbol != Invisible) {
      if ((m_symbol < Umbrella) || (m_symbol == Dot)) {
        painter->setBrush(painter->pen().color());
        painter->setPen(Qt::NoPen);
      }
      paintSymbol(painter, symbRect, m_symbol);
    }
  painter->translate(textRect().topLeft());
  m_doc.drawContents(painter);
}

QPixmap LegendItem::pixmap() const
{
  return KplItem::pixmap(m_pen, m_symbol);
}

void LegendItem::resetDashOffset()
{
  m_dashOffset = 0;
}

void LegendItem::setProperties(const QTransform& m)
{
  m_matrix = m;
  setProperties();
}

void LegendItem::setSize()
{
  QGraphicsItem* parent = parentItem();
  double r = (parent && parent->isVisible() &&
              parent->type() == FrameItem::Type) ?
              qgraphicsitem_cast<FrameItem*>(parent)->relSize() : 1;
  m_pen.setWidthF(r * penWidth(this) * m_symbolSize / m_relSize);
  QFont f("Helvetica");
  f.setPixelSize(qRound(0.033333 * parentWidth(this) * m_relSize));
  m_doc.setDefaultFont(f);
  if (m_symbol < SolidLine) {
    double dx = 0.01 * r * m_symbolSize * parentWidth(this);
    double dy = m_matrix.m22() / m_matrix.m11() * dx;
    symbRect.setCoords(-dx, dy, dx, -dy);
  }
}

void LegendItem::update()
{
  init();
  prepareGeometryChange();
  setProperties();
  QGraphicsItem::update();
}

void LegendItem::writePlo(KConfigGroup* plo) const
{
  plo->writeEntry("Type", "LEGENDITEM");
  DecoItem::writePlo(plo, isVisible());
  plo->writeEntry("symb", int(m_symbol));
  plo->writeEntry("coltext", QString::number(rgbQt1(m_textColor), 16));
  plo->writeEntry("xtext", m_textX);
  plo->writeEntry("ytext", m_textY);
  plo->writeEntry("symbsize", m_symbolSize);
  plo->writeEntry("s", m_text);
}

void LegendItem::init()
{
  TextItem::setText(m_doc, m_text, m_textColor);
  QTextFrameFormat tf = m_doc.rootFrame()->frameFormat();
  tf.setMargin(0);
  m_doc.rootFrame()->setFrameFormat(tf);
}

void LegendItem::mouseReleaseEvent(QGraphicsSceneMouseEvent* event)
{
  QGraphicsItem::mouseReleaseEvent(event);
  ((KplScene*)scene())->updatePos(this);
}

void LegendItem::setProperties()
{
  setPos(m_matrix.map(QPointF(m_x0, m_y0)));
  m_pen = createPen(this, m_color, DataPen, qMax(SolidLine, m_symbol));
  setSize();
}

QRectF LegendItem::textRect() const
{
  QRectF r = m_doc.documentLayout()->blockBoundingRect(m_doc.begin());
  r.moveTopLeft(r.topLeft() +
                QPointF(m_matrix.m11() * m_textX, m_matrix.m22() * m_textY -
                        QFontMetricsF(m_doc.defaultFont()).ascent()));
  return r;
}
