/***************************************************************************
                          lmfit.h  -  description
                             -------------------
    begin                : Sat May 22 1999
    copyright            : (C) 2015 by Werner Stille
    email                : stille@uni-freiburg.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef LMDIF_H
#define LMDIF_H

#include <QObject>
#include <kdemacros.h>

class QString;
class ArrayItem;
class FunItem;

/**
  * Parameter fit class. Provides methods for nonlinear Levenberg-Marquardt
  * parameter fits and general linear least square parameter fits.
  *
  * The functions lmdif, lmpar, fdjac2, qrsolv, qrfac, and enorm use code from
  * the public domain Fortran version of Argonne National Laboratories MINPACK,
  * C translation by Steve Moshier. Modifications by Werner Stille.
  *
  * The functions minv, igamc, igam, simq, and mtransp are from the Cephes Math
  * Library Release 2.7, Copyright by Stephen L. Moshier.
  *
  * The function lgam (original name: gsl_sf_lngamma_impl, Author: G. Jungman)
  * is from the GNU Scientific Library. Modifications by Werner Stille.
  */
class KDE_EXPORT LMFit : public QObject
{
  Q_OBJECT
public:
  /**
   * Constructor.
   * @param aList list of array items.
   * @param fList list of function items.
   * @param bFit array containing true values for parameters to be fitted.
   * @param par array of parameters.
   * @param sigma array for data errors.
   * @param parent pointer to parent widget.
   */
  LMFit(const QList<const ArrayItem*>* aList, QList<FunItem*>* fList,
        const bool* bFit, double* par, QObject* parent = 0,
        const double* sigma = 0);
  /** Destructor. */
  ~LMFit();
  /**
   * Calculates the euclidean norm of x.
   * @param n length of x.
   * @param x input array.
   * @return euclidean norm of x.
   */
  static double enorm(int n, const double x[]);
  /**
   * Calculates complemented incomplete gamma integral.
   * @param a argument of gamma function.
   * @param x upper boundary of integral.
   * @return complemented incomplete gamma integral.
   */
  static double igamc(double a, double x);
  /**
   * Performs general linear least square parameter fit.
   * @param n number of data points.
   * @param nPar number of parameters to be fitted.
   * @param par array of parameters to be fitted.
   * @param resVector array for weighted residuals.
   * @param fJac array for calculating the Jacobian matrix.
   * @param iPvt array defining a permutation matrix for calculating the
   * Jacobian matrix.
   * @return true in case of termination by user.
   */
  bool linFit(int n, int nPar, double par[], double resVector[], double fJac[],
              int iPvt[]);
  /**
   * Performs nonlinear Levenberg-Marquardt parameter fit.
   * @param n number of data points.
   * @param nPar number of parameters to be fitted.
   * @param par array of parameters to be fitted.
   * @param resVector array for weighted residuals.
   * @param tol tolerance value for stopping nonlinear fit iterations.
   * @param itMax maximum number of nonlinear fit iterations.
   * @param info error code.
   * @param fJac array for calculating the Jacobian matrix.
   * @param iPvt array defining a permutation matrix for calculating the
   * Jacobian matrix.
   */
  void lmdif(int n, int nPar, double par[], double resVector[], double tol,
             int itMax, int* info, double fJac[], int iPvt[]);
  /**
   * Finds the inverse of a square matrix.
   * @param a input matrix.
   * @param x output matrix.
   * @param n numbers of rows and columns of input matrix.
   * @return 0 for success.
   */
  static int minv(double a[], double x[], int n);
  /**
   * Sets data error array.
   * @param sigma array for data errors.
   */
  inline void setSigma(const double* sigma) { m_sigma = sigma; };
  /**
   * Sets condition to stop calculations.
   * @param userBreak true to break calculations.
   */
  inline void setUserBreak(bool userBreak = true) { m_userBreak = userBreak; };
  /** Returns current break condition for calculations. */
  inline bool userBreak() const { return m_userBreak; };

protected:
  void fcn(int n, int np, const double p[], double fvec[], int* iflag);
  void fdjac2(int m, int n, double x[], const double fvec[], double fjac[],
              int ldfjac, int* iflag, double epsfcn, double wa[]);
  void lmdif(int m, int np, double px[], double fvec[], double ftol,
             double ptol, double gtol, int maxfev, double epsfcn,
             double diag[], int mode, double factor, int nprint,
             int* info, int* nfev, double fjac[], int ldfjac,
             int ipvt[], double qtf[], double wa1[], double wa2[],
             double wa3[], double wa4[]);
  bool m_userBreak;
  const bool* m_bFit;
  double* m_par;
  const double* m_sigma;
  const QList<const ArrayItem*>* m_aList;
  QList<FunItem*>* m_fList;

signals:
  /**
    * Emitted when a message should be displayed.
    * @param message message text
    */
  void updateMessage(const QString& message);
};

#endif
