/***************************************************************************
                          parfunitem.cpp  -  description
                             -------------------

    This file is a part of kpl - a program for graphical presentation of
    data sets and functions.

    begin                : Sun Aug 29 1999
    copyright            : (C) 2015 by Werner Stille
    email                : stille@uni-freiburg.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <QFileInfo>
#include <QLibrary>
#include <QPainter>
#include <QProcess>
#include <QTextStream>
#include <KConfigGroup>
#include <KLocalizedString>
#include <KMessageBox>
#include <ThreadWeaver/JobCollection>
#include <ThreadWeaver/Weaver>
#include "arrayitem.h"
#include "calcjob.h"
#include "frameitem.h"
#include "parfunitem.h"
#include "kplscene.h"

ParFunItem::ParFunItem(const KUrl& urlX, const QString& nameX,
                       const KUrl& urlY, const QString& nameY, unsigned color,
                       Symbol symbol, double fx, double fy,
                       QGraphicsItem* parent, double tMin, double tMax,
                       double deltaT, bool visible, FillStyle fillStyle,
                       double xOffset, double yOffset, double relSize) :
 FunItem(urlY, nameY, color, symbol, fx, fy, parent, tMin, tMax, deltaT,
         visible, fillStyle, xOffset, yOffset, relSize),
 funcXo(0), m_nameX(nameX), m_urlX(urlX), m_libX(0)
{
  init();
  if (!m_urlX.url().isEmpty())
    getFuncAddr(m_urlX.path(), m_nameX, &m_libX, &funcX);
}

ParFunItem::ParFunItem(const ParFunItem& a) :
 FunItem(a), funcX(a.funcX), funcXo(a.funcXo), m_xValues(a.m_xValues),
 m_nameX(a.m_nameX), m_urlX(a.m_urlX), m_libX(0)
{
  memcpy(m_parX, a.m_parX, sizeof(m_parX));
  memcpy(parXo, a.parXo, sizeof(parXo));
  if (a.m_libX)
    getFuncAddr(m_urlX.path(), m_nameX, &m_libX, &funcX);
}

ParFunItem::ParFunItem(const KConfigGroup* plo, const KUrl& urlPlo,
                       unsigned color, Symbol symbol, QGraphicsItem* parent) :
 funcXo(0)
{
  setParentItem(parent);
  ScaledItem::readPlo(plo, symbol, color);
  m_tMin = plo->readEntry("tmin", 0.0);
  m_tMax = plo->readEntry("tmax", m_tMin + 1);
  m_deltaT = plo->readEntry("dt", 0.0);
  stringToPar(plo->readEntry("px"), m_parX);
  stringToPar(plo->readEntry("py"), m_parY);
  m_nameX = plo->readEntry("namex", "");
  m_nameY = plo->readEntry("namey", "");
  m_urlX = absoluteUrl(plo->readEntry("pathx", ""), urlPlo);
  m_urlY = absoluteUrl(plo->readEntry("pathy", ""), urlPlo);
  getFuncAddr(m_urlX.path(), m_nameX, &m_libX, &funcX);
  getFuncAddr(m_urlY.path(), m_nameY, &m_libY, &funcY);
}

ParFunItem::~ParFunItem()
{
  unloadX();
}

QRectF ParFunItem::boundingRect() const
{
  if (isVisible() && bRect.isNull())
    paintCond(0);
  return bRect;
}

int ParFunItem::calcTable() const
{
  bool newv = m_tValues.isEmpty();
  double d;
  int n;
  if (m_deltaT) {
    n = int(fabs((m_tMax - m_tMin) / m_deltaT)) + 1;
    d = m_deltaT;
  } else {
    n = 101;
    d = 0.01 * (m_tMax - m_tMin);
  }
  if (newv || (m_tMin != tMino) || (m_tMax != tMaxo) || (d != deltaTo)) {
    m_tValues.resize(n);
    int n1 = n - 1;
    for (int i = 0; i < n; i++)
      m_tValues[i] = (i == n1) ? m_tMax : (m_tMin + i * d);
  }
  bool pchanged = parChanged(m_parY, parYo);
  if (pchanged || newv || (m_tMin != tMino) || (m_tMax != tMaxo) ||
      (d != deltaTo) || (funcY != funcYo))
    m_yValues.resize(n);
  if (pchanged || newv || (m_tMin != tMino) || (m_tMax > tMaxo) ||
      (d != deltaTo) || (funcY != funcYo))
    FunItem::calcTable(m_tValues.data(), m_yValues.data(), funcY, m_parY, n);
  if (pchanged)
    memcpy(parYo, m_parY, sizeof(parYo));
  pchanged = parChanged(m_parX, parXo);
  if (pchanged || newv || (m_tMin != tMino) || (m_tMax != tMaxo) ||
      (d != deltaTo) || (funcX != funcXo))
    m_xValues.resize(n);
  if (pchanged || newv || (m_tMin != tMino) || (m_tMax > tMaxo) ||
      (d != deltaTo) || (funcX != funcXo))
    FunItem::calcTable(m_tValues.data(), m_xValues.data(), funcX, m_parX, n);
  if (pchanged)
    memcpy(parXo, m_parX, sizeof(parXo));
  funcXo = funcX;
  funcYo = funcY;
  tMino = m_tMin;
  tMaxo = m_tMax;
  deltaTo = d;
  return n;
}

void ParFunItem::exponent(int& expX, int& expY, double& fx, double& fy) const
{
  if (funcX && funcY) {
    double xmin, xmax, ymin, ymax;
    limits(xmin, xmax, ymin, ymax);
    ScaledItem::exponent(qMax(fabs(xmin), fabs(xmax)), expX, fx);
    ScaledItem::exponent(qMax(fabs(ymin), fabs(ymax)), expY, fy);
  }
}

void ParFunItem::exportTable(QTextStream& stream, char separator, char format,
                             int precision) const
{
  int n = calcTable();
  for (int i = 0; i < n; i++)
    stream << QString::number(m_normX * m_xValues[i], format, precision)
           << separator
           << QString::number(m_normY * m_yValues[i], format, precision)
           << "\n";
}

void ParFunItem::limits(double& xMin, double& xMax, double& yMin,
                        double& yMax) const
{
  if (funcX && funcY) {
    int n = calcTable();
    ScaledItem::limits(xMin, xMax, m_xValues.data(), n);
    ScaledItem::limits(yMin, yMax, m_yValues.data(), n);
    xMin += m_xOffset;
    xMax += m_xOffset;
    yMin += m_yOffset;
    yMax += m_yOffset;
  }
}

void ParFunItem::paint(QPainter* painter, const QStyleOptionGraphicsItem*,
                       QWidget*)
{
  paintCond(painter);
}

void ParFunItem::setDashOffset()
{
  if (!m_fillStyle  && (m_symbol > 1)) {
    m_pen.setDashOffset(m_dashOffset);
    double s = 0;
    foreach (double d, m_pen.dashPattern())
      s += d;
    m_dashOffset += length(m_xValues.data(), m_yValues.data(), calcTable()) /
      m_pen.widthF();
    m_dashOffset -= int(m_dashOffset / s) * s;
  }
}

bool ParFunItem::setFuncX(const KUrl& url, const QString& name)
{
  m_urlX = url;
  m_nameX = name;
  unloadX();
  return getFuncAddr(url.path(), name, &m_libX, &funcX);
}

void ParFunItem::setPar(int i, double value, bool yFun)
{
  if (yFun)
    m_parY[i] = value;
  else
    m_parX[i] = value;
}

void ParFunItem::writePlo(KConfigGroup* plo, const KUrl& urlPlo, bool absolute,
                          char format, int precision) const
{
  plo->writeEntry("Type", "PARFUNITEM");
  ScaledItem::writePlo(plo, isVisible());
  plo->writeEntry("tmin", QString::number(m_tMin, format, precision));
  plo->writeEntry("tmax", QString::number(m_tMax, format, precision));
  plo->writeEntry("dt", QString::number(m_deltaT, format, precision));
  plo->writeEntry("px", parToString(m_parX, format, precision));
  plo->writeEntry("py", parToString(m_parY, format, precision));
  plo->writeEntry("namex", m_nameX);
  plo->writeEntry("namey", m_nameY);
  plo->writeEntry("pathx", path(urlPlo, m_urlX, absolute));
  plo->writeEntry("pathy", path(urlPlo, m_urlY, absolute));
}

double* ParFunItem::xValues() const
{
  return m_xValues.data();
};

void ParFunItem::init()
{
  memset(m_parX, 0, sizeof(m_parX));
  memset(parXo, 0, sizeof(parXo));
}

void ParFunItem::paintCond(QPainter* painter) const
{
  if (!painter)
    bRect = QRectF();
  if (!funcY || !funcX || m_symbol == -17)
    return;
  FrameItem* frame = qgraphicsitem_cast<FrameItem*>(parentItem());
  bool useFrame = frame && frame->isVisible();
  if (painter) {
    painter->setClipRect(useFrame ?
                         frame->clipRect() :
                         static_cast<KplScene*>(scene())->sceneRect());
    if (m_fillStyle) {
      painter->setBrush(createBrush(m_color, m_fillStyle));
      painter->setPen(Qt::NoPen);
    } else
      painter->setPen(m_pen);
  }
  plotArray(painter, m_xValues.data(), m_yValues.data(), calcTable());
}

void ParFunItem::unloadX()
{
  if (m_libX) {
    m_libX->unload();
    delete m_libX;
    m_libX = 0;
  }
}
