//
// C++ Interface: psgraph
//
// Description: 
//
//
// Author: Werner Stille <stille@uni-freiburg.de>, (C) 2015
//
// Copyright: See COPYING file that comes with this distribution
//
//
#ifndef PSGRAPH_H
#define PSGRAPH_H

#include <QPointF>
#include <QSizeF>
#include "textitem.h"

class QGraphicsItem;
class QTextStream;
class ArcItem;
class ArrayItem;
class Array3DItem;
class ArrowItem;
class EllipseItem;
class FrameItem;
class FunItem;
class Fun3DItem;
class ImageItem;
class LegendItem;
class LineItem;
class ParFunItem;
class RectItem;
class ScaleBarItem;
class ScaledItem;
class SplineItem;
class TextItem;
class ThreeDItem;

/**
  * PS graphics class.
  * Provides methods to generate PS and EPS output.
  * @author Werner Stille
  */
class KDE_EXPORT PSGraph
{
public:
  /**
   * Constructor. Creates PS or EPS output of plot items.
   * @param stream pointer to output stream.
   * @param list list of plot items.
   * @param size media size in cm.
   * @param landscape true for landscape orientation.
   * @param eps true for EPS, false for PS output.
   */
  PSGraph(QTextStream* stream, const QList<QGraphicsItem*>& list,
          const QSizeF& size, bool landscape = false, bool eps = true);
  /**
   * Constructor. Writes PS prolog.
   * @param stream pointer to output stream.
   * @param size media size in cm.
   * @param topRight top-right coordinates of graphics rectangle in cm.
   * @param bottomLeft bottom-left coordinates of graphics rectangle in cm.
   * @param landscape true for landscape orientation.
   * @param eps true for EPS, false for PS output.
   */
  PSGraph(QTextStream* stream, const QSizeF& size,
          const QPointF& topRight = QPointF(),
          const QPointF& bottomLeft = QPointF(), bool landscape = false,
          bool eps = true);
  /** Destructor. */
  ~PSGraph();
  /**
   * Plots array item.
   * @param x array with x values.
   * @param y array with y values.
   * @param fx normalization factor for x values.
   * @param fy normalization factor for y values.
   * @param n number of data points.
   * @param fillStyle fill style.
   * @param xOff offset for x values.
   * @param yOff offset for x values.
   * @param clip true for clipping.
   */
  void drawArray(const double* x, const double* y, double fx, double fy, int n,
                 KplItem::FillStyle fillStyle = KplItem::NoBrush,
                 double xOff = 0, double yOff = 0, bool clip = true);
  /**
   * Plots items.
   * @param list list of items.
   */
  void drawItems(const QList<QGraphicsItem*>& list);
  /**
   * Plots text.
   * @param text text.
   * @param x x value in points.
   * @param y y value in points.
   * @param alignment horizontal alignment.
   */
  void drawText(const QString& text, double x, double y,
                TextItem::Alignment alignment = TextItem::AlignLeft);
  /** Finishes plot. */
  void finish();
  /**
   * Sets color for data plotting.
   * @param color color.
   */
  void setColorData(unsigned color);
  /**
    * Sets rotation angle.
    * @param dir rotation angle in deg.
    */
  void setDir(double dir);
  /**
    * Sets relative size of item.
    * @param m_relSize relative size.
    * @return old relative size.
    */
  double setRelSize(double m_relSize);
  /** Points per cm. */
  static const double pts = 72 / 2.54;

protected:
  /**
   * Converts from centimeters to user units.
   * @param xcm x value in centimeters.
   * @param ycm y value in centimeters.
   * @param fx normalization factor for x values.
   * @param fy normalization factor for y values.
   * @param xu pointer to x value in user units (output).
   * @param yu pointer to y value in user units (output).
   */
  void cm2r(double xcm, double ycm, double fx, double fy, double* xu,
            double* yu) const;
  /**
   * Plots 3D item.
   * @param item pointer to item.
   */
  void draw3D(ThreeDItem* item);
  /**
   * Plots arc item.
   * @param item pointer to item.
   */
  void drawArc(ArcItem* item);
  /**
   * Plots array item.
   * @param item pointer to item.
   */
  void drawArray(ArrayItem* item);
  /**
   * Plots 3D array item.
   * @param item pointer to item.
   */
  void drawArray3D(Array3DItem* item);
  /**
   * Plots arrow item.
   * @param item pointer to item.
   */
  void drawArrow(ArrowItem* item);
  /**
   * Plots ellipse item.
   * @param item pointer to item.
   */
  void drawEllipse(EllipseItem* item);
  /**
   * Plots error bars for array item.
   * @param x array with x values.
   * @param y array with y values.
   * @param errors array with y error values.
   * @param fx normalization factor for x values.
   * @param fy normalization factor for y values.
   * @param n number of data points.
   * @param xOff offset for x values.
   * @param yOff offset for x values.
   * @param clip true for clipping.
   */
  void drawErrorBars(const double* x, const double* y, const double* errors,
                     double fx, double fy, int n, double xOff = 0,
                     double yOff = 0, bool clip = true);
  /**
   * Plots frame item.
   * @param item pointer to item.
   */
  void drawFrame(FrameItem* item);
  /**
   * Plots function item.
   * @param item pointer to item.
   */
  void drawFunc(FunItem* item);
  /**
   * Plots 3D function item.
   * @param item pointer to item.
   */
  void drawFunc3D(Fun3DItem* item);
  /**
   * Plots image item.
   * @param item pointer to item.
   */
  void drawImage(ImageItem* item);
  /**
   * Plots legend item.
   * @param item pointer to item.
   */
  void drawLegend(LegendItem* item);
  /**
   * Plots line item.
   * @param item pointer to item.
   */
  void drawLine(LineItem* item);
  /**
   * Plots parametric function item.
   * @param item pointer to item.
   */
  void drawParFunc(ParFunItem* item);
  /**
   * Plots rectangle item.
   * @param item pointer to item.
   */
  void drawRect(RectItem* item);
  /**
   * Plots scale bar item.
   * @param item pointer to item.
   */
  void drawScaleBar(ScaleBarItem* item);
  /**
   * Plots spline item.
   * @param item pointer to item.
   */
  void drawSpline(SplineItem* item);
  /**
   * Plots text item.
   * @param item pointer to item.
   */
  void drawText(TextItem* item);
  /**
   * Plots x axis label.
   * @param text label text.
   * @param exp logarithm of normalization factor for x axis.
   * @param xTop true for top side, false for bottom side label.
   */
  void labelX(const QString& text, int exp, bool xTop);
  /**
   * Plots y axis label.
   * @param text label text.
   * @param exp logarithm of normalization factor for y axis.
   * @param yRight true for right side, false for left side label.
   */
  void labelY(const QString& text, int exp, bool yRight);
  /**
   * Writes PostScript prolog.
   * @param landscape true for landscape orientation.
   * @param eps true for EPS, false for PS output.
   * @param topRight coordinates of top-right corner in cm.
   * @param bottomLeft coordinates of bottom-left corner in cm.
   */
  void prolog(bool landscape, bool eps,
              const QPointF& topRight = QPointF(),
              const QPointF& bottomLeft = QPointF());
  /**
   * Converts to PostScript string.
   * @param text text.
   */
  void psStr(const QString& text);
  /** Resets offset for dashed lines. */
  void resetDashOffset();
  /**
   * Scales data window to user units.
   * @param xMin x value of left data window margin in user units.
   * @param xMax x value of right data window margin in user units.
   * @param yMin y value of bottom data window margin in user units.
   * @param yMax y value of top window data margin in user units.
   * @param logX true for logarithmic x scale.
   * @param logY true for logarithmic y scale.
   */
  void scale(double xMin, double xMax, double yMin, double yMax,
             bool logX = false, bool logY = false);
  /**
   * Sets color for frame.
   * @param color color.
   */
  void setColorFrame(unsigned color);
  /**
   * Sets color for grid and tics.
   * @param color color.
   */
  void setColorGrid(unsigned color);
  /**
   * Sets color and symbol type for array and function items.
   * @param item pointer to item.
   */
  void setProperties(ScaledItem* item);
  /**
   * Sets symbol type.
   * @param symbol symbol type.
   */
  void setSymbol(KplItem::Symbol symbol);
  /**
   * Sets position and size of the data window.
   * @param x0 left page margin in centimeters.
   * @param width width in centimeters.
   * @param y0 bottom page margin in centimeters.
   * @param height height in centimeters.
   */
  void window(double x0, double width, double y0, double height);
  bool m_logX, m_logY;
  KplItem::Symbol m_symbol;
  double m_x0, m_y0, m_width, m_height, m_xMin, m_yMin, m_xMax, m_yMax,
    m_relSize;
  QTextStream* ts;
  QSizeF m_size;
};

#endif
