/***************************************************************************
                          scalebaritem.h  -  description
                             -------------------
    begin                : Don Jun 12 2003
    copyright            : (C) 2015 by Werner Stille
    email                : stille@uni-freiburg.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef SCALEBARITEM_H
#define SCALEBARITEM_H

#include <QGraphicsItem>
#include <QTextDocument>
#include "decoitem.h"

/**
  * Scale bar item class. Class for scale bar decoration items.
  * @author Werner Stille
  */
class KDE_EXPORT ScaleBarItem : public DecoItem, public QGraphicsItem
{
public:
  /** Specific QGraphicsItem type number for function type() */
  enum {Type = UserType + ScaleBar
    /**< Specific QGraphicsItem type number of ScaleBarItem for function
     * type().
     */
  };
  /**
    * Constructor. Initializes corresponding to arguments.
    * @param x0 x position of bar in cm.
    * @param y0 y position of bar in cm.
    * @param color color for scale bar.
    * @param parent pointer to parent item.
    * @param length length in cm.
    * @param text text.
    * @param visible true for visible item.
    * @param vertical false for horizontal, true for vertical orientation.
    * @param relSize relative size of scale bar.
    */
  ScaleBarItem(double x0 = 0, double y0 = 0, unsigned color = 0,
               QGraphicsItem* parent = 0, double length = 1,
               const QString& text = "", bool visible = true,
               bool vertical = false, double relSize = 1);
  /**
    * Copy constructor.
    * @param a source item.
    */
  ScaleBarItem(const ScaleBarItem& a);
  /**
    * Constructor. Initializes corresponding to plot file.
    * @param plo pointer to KSimpleConfig object containing item data.
    * @param color default color for scale bar.
    * @param parent pointer to parent item.
    */
  ScaleBarItem(const KConfigGroup* plo, unsigned color = 0,
               QGraphicsItem* parent = 0);
  /** Destructor. */
  virtual ~ScaleBarItem();
  /** Calculates bounding rectangle. */
  virtual QRectF boundingRect() const;
  /** Returns length in cm. */
  inline double length() const { return m_length; };
  /** Paints item. */
  virtual void paint(QPainter* painter, const QStyleOptionGraphicsItem* option,
                     QWidget* widget = 0);
  /** Returns decoration pixmap for item list. */
  QPixmap pixmap() const;
  /**
    * Sets length.
    * @param length length in cm.
    */
  inline void setLength(double length) { m_length = length; };
  /**
    * Sets properties according to transformation matrix.
    * @param m transformation matrix.
    */
  void setProperties(const QTransform& m);
  /** Sets size according to transformation matrix. */
  void setSize();
  /**
   * Sets text.
   * @param text text.
   */
  inline void setText(const QString& text) { m_text = text; };
  /**
   * Sets orientation.
   * @param vertical true for vertical, false for horizontal.
   */
  inline void setVertical(bool vertical) { m_vertical = vertical; };
  /** Returns text. */
  inline QString text() const { return m_text; };
  /** Returns specific QGraphicsItem type number. */
  virtual int type() const { return Type; };
  /** Applies changes and forces redrawing. */
  void update();
  /** Returns true for vertical, false for horizontal orientation. */
  inline bool vertical() const { return m_vertical; };
  /**
   * Writes into plot file.
   * @param plo KConfigGroup object.
   */
  void writePlo(KConfigGroup* plo) const;

protected:
  /** Performs initialization. */
  void init();
  /** Updates position. */
  virtual void mouseReleaseEvent(QGraphicsSceneMouseEvent* event);
  /** Sets properties according to transformation matrix. */
  void setProperties();
  /** Returns text rectangle in pixel units. */
  QRectF textRect() const;
  bool m_vertical;
  double m_length, l2, wRel;
  QString m_text;
  QTextDocument m_doc;
  QPen m_pen;
};

#endif
