/***************************************************************************
                          splineitem.cpp  -  description
                             -------------------
    begin                : Wed Feb 27 2002
    copyright            : (C) 2015 by Werner Stille
    email                : stille@uni-freiburg.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <QLibrary>
#include <QPainter>
#include <QTextStream>
#include <KConfigGroup>
#include "arrayitem.h"
#include "fitpack.h"
#include "frameitem.h"
#include "kplscene.h"
#include "splineitem.h"

/**
 * Compares table rows for qsort.
 * @param e1 pointer to row 1.
 * @param e2 pointer to row 2.
 * @return -1 if x1 < x2, 0 if x1 = x2, 1 if x1 > x2.
 */
static int cmpasc(const void *e1, const void *e2)
{
  double a1 = *((double*) e1);
  double a2 = *((double*) e2);
  if (a1 < a2)
    return -1;
  if (a1 > a2)
    return 1;
  return 0;
}

SplineItem::SplineItem(int degree, int derivOrder, double xLow, unsigned color,
                       Symbol symbol, double fx, double fy,
                       QGraphicsItem* parent, double xMin, double xMax,
                       double deltaX, bool visible, FillStyle fillStyle,
                       double xOffset, double yOffset, double relSize) :
 ScaledItem(fillStyle, symbol, color, fx, fy, xOffset, yOffset, relSize,
            parent),
 logXo(false), m_degree(degree), m_numKnots(0), m_derivOrder(derivOrder),
 degreeo(degree), numKnotso(0), derivOrdero(0), m_xMin(xMin), m_xMax(xMax),
 m_deltaX(deltaX), m_xLow(xLow), xMino(0), xMaxo(0), deltaXo(0), xLowo(0)
{
  if (!visible)
    hide();
}

SplineItem::SplineItem(const SplineItem& a) :
 ScaledItem(a), logXo(a.logXo), m_degree(a.m_degree), m_numKnots(a.m_numKnots),
 m_derivOrder(a.m_derivOrder), degreeo(a.degreeo), numKnotso(a.numKnotso),
 derivOrdero(a.derivOrdero), m_xMin(a.m_xMin), m_xMax(a.m_xMax),
 m_deltaX(a.m_deltaX), m_xLow(a.m_xLow), xMino(a.xMino), xMaxo(a.xMaxo),
 deltaXo(a.deltaXo), xLowo(a.xLowo), m_xValues(a.m_xValues),
 m_yValues(a.m_yValues), t(a.t), c(a.c)
{
}

SplineItem::SplineItem(const KConfigGroup* plo, unsigned color, Symbol symbol,
                       QGraphicsItem* parent) :
 ScaledItem(plo, symbol, color, parent), logXo(false), degreeo(3),
 numKnotso(0), derivOrdero(0), xMino(0), xMaxo(0), deltaXo(0), xLowo(0)
{
  m_degree = plo->readEntry("kdeg", 3);
  m_numKnots = plo->readEntry("nk", 0);
  m_derivOrder = plo->readEntry("nderiv", 0);
  m_xMin = plo->readEntry("xmin", 0.0);
  m_xMax = plo->readEntry("xmax", m_xMin + 1);
  m_deltaX = plo->readEntry("dx", 0.0);
  m_xLow = plo->readEntry("xlow", m_xMin);
  t.resize(m_numKnots);
  for (int i = 0; i < m_numKnots; i++)
    t[i] = plo->readEntry(QString("t%1").arg(i), 0.0);
  int nc = qMax(0, m_numKnots - m_degree - 1);
  c.resize(nc);
  for (int i = 0; i < nc; i++)
    c[i] = plo->readEntry(QString("c%1").arg(i), 0.0);
}

SplineItem::~SplineItem()
{
}

QRectF SplineItem::boundingRect() const
{
  if (isVisible() && bRect.isNull())
    paintCond(0);
  return bRect;
}

int SplineItem::calcTable(bool logX) const
{
  bool newv = m_xValues.isEmpty();
  double d;
  int n;
  if (m_deltaX) {
    n = int(fabs((m_xMax - m_xMin) / m_deltaX)) + 1;
    d = logX ? (log10(m_xMax / m_xMin) / (n - 1)) : m_deltaX;
  } else {
    n = 101;
    d = 0.01 * (logX ? log10(m_xMax / m_xMin) : (m_xMax - m_xMin));
  }
  int i;
  if (newv || (logX != logXo) || (m_degree != degreeo) ||
      (m_numKnots != numKnotso) || (m_derivOrder != derivOrdero) ||
      (m_xMin != xMino) || (m_xMax != xMaxo) || (d != deltaXo) ||
      (m_xLow != xLowo)) {
    m_xValues.resize(n);
    m_yValues.resize(n);
    int n1 = n - 1;
    for (i = 0; i < n; i++)
      m_xValues[i] = (i == n1) ? m_xMax :
                                 (logX ? pow(10, log10(m_xMin) + i * d) :
                                         (m_xMin + i * d));
    calcTable(m_xValues.data(), m_yValues.data(), n, m_derivOrder, m_xLow);
  }
  logXo = logX;
  degreeo = m_degree;
  numKnotso = m_numKnots;
  derivOrdero = m_derivOrder;
  xMino = m_xMin;
  xMaxo = m_xMax;
  deltaXo = d;
  xLowo = m_xLow;
  return n;
}

int SplineItem::calcTable(const double* x, double* values, int n,
                          int derivOrder, double xLow) const
{
  int ier = 0;
  if (derivOrder >= 0)
    FitPack::splder(t.data(), m_numKnots, c.data(), m_degree, derivOrder, x,
                    values, n, &ier);
  else
    for (int i = 0; i < n; i++)
      values[i] = FitPack::splint(t.data(), m_numKnots, c.data(), m_degree,
                                  xLow, x[i]);
  return ier;
}

void SplineItem::copySpline(const SplineItem* a)
{
  t = a->t;
  c = a->c;
  m_numKnots = a->numKnots();
  m_degree = a->m_degree;
  m_xValues.resize(0);
};

void SplineItem::exponent(int& expX, int& expY, double& fx, double& fy,
                          bool logX) const
{
  if (m_numKnots) {
    ScaledItem::exponent(qMax(fabs(m_xMin), fabs(m_xMax)), expX, fx);
    double xmi, xma, ymi, yma;
    limits(xmi, xma, ymi, yma, logX);
    ScaledItem::exponent(qMax(fabs(ymi), fabs(yma)), expY, fy);
  }
}

void SplineItem::exportTable(QTextStream& stream, char separator, char format,
                             int precision) const
{
  int n = calcTable(logXo);
  for (int i = 0; i < n; i++)
    stream << QString::number(m_normX * m_xValues[i], format, precision)
           << separator
           << QString::number(m_normY * m_yValues[i], format, precision)
           << "\n";
}

int SplineItem::fit(const ArrayItem* ai, double xMin, double xMax,
                    double* chiSquare, int* numKnots, double smooth,
                    FunItem::DataErrorStruct* err, double* avgErr)
{
  return fit(&ai->data()[ai->colX()][ai->rowStart()],
             &ai->data()[ai->colY()][ai->rowStart()],
             &ai->data()[ai->colErr()][ai->rowStart()], ai->count(),
             xMin, xMax, chiSquare, numKnots, smooth, err, avgErr);
}

int SplineItem::fit(const double* x0, const double *y0, const double* errors,
                    int n, double xMin, double xMax, double* chiSquare,
                    int* numKnots, double smooth,
                    FunItem::DataErrorStruct* err, double* avgErr)
{
  QVector<double> x(n);
  QVector<double> y(n);
  QVector<double> w(n);
  int ier = sortArrays(x0, y0, errors, n, err, x.data(), y.data(), w.data());
  if (ier)
    return ier;
  if (smooth)
    for (int i = 0; i < n; i++)
      w[i] = 1 / w[i];
  int nest = n + m_degree + 1;
  QVector<double> tf(nest);
  QVector<double> cf(nest);
  int nkt;
  double chisqt;
  FitPack::curfit(n, x.data(), y.data(), w.data(), xMin, xMax, m_degree,
                  smooth * smooth * n, nest, &nkt, tf.data(), cf.data(),
                  &chisqt, &ier);
  if (ier < 4) {
    m_xMin = qMax(m_xMin, xMin);
    m_xMax = qMin(m_xMax, xMax);
    m_numKnots = nkt;
    if (numKnots)
      *numKnots = nkt;
    if (chiSquare)
      *chiSquare = chisqt;
    t.resize(m_numKnots);
    qCopy(tf.begin(), tf.begin() + m_numKnots, t.begin());
    int nc = m_numKnots - m_degree - 1;
    c.resize(nc);
    qCopy(cf.begin(), cf.begin() + nc, c.begin());
    m_xValues.resize(0);
    if (avgErr) {
      *avgErr = 0;
      if (smooth) {
        QVector<double> fv(n);
        FitPack::splev(t.data(), m_numKnots, c.data(), m_degree, x.data(),
                       fv.data(), n, &ier);
        for (int i = 0; i < n; i++) {
          double e = y[i] - fv[i];
          *avgErr += e * e;
        }
        *avgErr = sqrt(*avgErr / n);
      }
    }
  }
  return ier;
}

int SplineItem::interpolation(const double* x, const double* y, int n,
                              double xMin, double xMax)
{
  QVector<double> e(n);
  e.fill(1.0);
  int nkt;
  double chisq;
  return fit(x, y, e.data(), n, xMin, xMax, &chisq, &nkt);
}

void SplineItem::limits(double& xMin, double& xMax, double& yMin,
                        double& yMax, bool logX) const
{
  if (m_numKnots) {
    xMin = m_xMin + m_xOffset;
    xMax = m_xMax + m_xOffset;
    int n = calcTable(logX);
    ScaledItem::limits(yMin, yMax, m_yValues.data(), n);
    yMin += m_yOffset;
    yMax += m_yOffset;
  }
}

void SplineItem::paint(QPainter* painter, const QStyleOptionGraphicsItem*,
                       QWidget*)
{
  paintCond(painter);
}

int SplineItem::roots(QVector<double>& zero, int derivOrder,
                      double offset) const
{
  int mest = 3 * m_numKnots;
  zero.resize(mest);
  int ier;
  int nz;
  FitPack::sproot(t.data(), m_numKnots, c.data(), zero.data(), mest, &nz, &ier,
                  derivOrder, offset);
  zero.resize(nz);
  return ier;
}

void SplineItem::setDashOffset()
{
  if (!m_fillStyle  && (m_symbol > 1)) {
    m_pen.setDashOffset(m_dashOffset);
    double s = 0;
    foreach (double d, m_pen.dashPattern())
      s += d;
    FrameItem* frame = qgraphicsitem_cast<FrameItem*>(parentItem());
    m_dashOffset += length(m_xValues.data(), m_yValues.data(),
                           calcTable((frame && frame->isVisible()) ?
                                     frame->logX() : false)) / m_pen.widthF();
    m_dashOffset -= int(m_dashOffset / s) * s;
  }
}

QPainterPath SplineItem::shape() const
{
  if (isVisible() && bPath.isEmpty())
    paintCond(0);
  return bPath;
}

int SplineItem::sortArrays(const double* x0, const double* y0,
                           const double* errors0, int& n,
                           FunItem::DataErrorStruct* err, double* x,
                           double* y, double* errors)
{
  int ier = 0;
  double (*fktErrMod)(double, const double*);
  QLibrary* lib;
  bool errcol = true;
  if (err) {
    errcol = err->fitErrCol;
    if (!errcol) {
      if (!FunItem::getFuncAddr(err->errModPath, err->errModName, &lib,
                                &fktErrMod))
        return 5;
    }
  }
  int nc = errcol ? 3 : 2;
  QVector<double> tmp(nc * n);
  int j = 0;
  for (int i = 0; i < n; i++) {
    tmp[j++] = x0[i];
    tmp[j++] = y0[i];
    if (errcol)
      tmp[j++] = errors0[i];
  }
  qsort(tmp.data(), n, nc * sizeof(double), cmpasc);
  j = 0;
  int ni = 0;
  for (int i = 0; i < n; i++) {
    x[ni] = tmp.at(j++);
    y[ni] = tmp.at(j++);
    if (errcol)
      errors[ni] = tmp.at(j++);
    else
      errors[ni] = FrameItem::valid(x[ni]) ?
                   fktErrMod(err->errModArg ? y[ni] : x[ni], err->parErrMod) :
                   NAN;
    if ((FrameItem::valid(x[ni])) && (FrameItem::valid(y[ni])) &&
        (FrameItem::valid(errors[ni])))
      ni++;
  }
  n = ni;
  for (int i = 0; i < n; i++) {
    if (errors[i] <= 0) {
      ier = 4;
      break;
    }
  }
  if (!errcol) {
    lib->unload();
    delete lib;
  }
  return ier;
}

void SplineItem::writePlo(KConfigGroup* plo, char format, int precision) const
{
  plo->writeEntry("Type", "SPLINEITEM");
  ScaledItem::writePlo(plo, isVisible());
  plo->writeEntry("kdeg", m_degree);
  plo->writeEntry("nk", m_numKnots);
  plo->writeEntry("nderiv", m_derivOrder);
  plo->writeEntry("xmin", QString::number(m_xMin, format, precision));
  plo->writeEntry("xmax", QString::number(m_xMax, format, precision));
  plo->writeEntry("dx", QString::number(m_deltaX, format, precision));
  plo->writeEntry("xlow", QString::number(m_xLow, format, precision));
  for (int i = 0; i < m_numKnots; i++)
    plo->writeEntry(QString("t%1").arg(i), t[i]);
  for (int i = 0; i < (m_numKnots - m_degree - 1); i++)
    plo->writeEntry(QString("c%1").arg(i), c[i]);
}

double* SplineItem::xValues() const
{
  return m_xValues.data();
};

double* SplineItem::yValues() const
{
  return m_yValues.data();
};

void SplineItem::paintCond(QPainter* painter) const
{
  if (!m_numKnots)
    return;
  FrameItem* frame = qgraphicsitem_cast<FrameItem*>(parentItem());
  KplScene* sc = 0;
  bool logx = false;
  bool useFrame = frame && frame->isVisible();
  if (useFrame)
    logx = frame->logX();
  else
    sc = static_cast<KplScene*>(scene());
  if (painter) {
    painter->setClipRect(useFrame ? frame->clipRect() : sc->sceneRect());
    if (m_fillStyle) {
      painter->setBrush(createBrush(m_color, m_fillStyle));
      painter->setPen(Qt::NoPen);
    } else
      painter->setPen(m_pen);
  }
  plotArray(painter, m_xValues.data(), m_yValues.data(), calcTable(logx));
}
