/***************************************************************************
                          textitem.h  -  description
                             -------------------
    begin                : Wed Apr 18 2001
    copyright            : (C) 2015 by Werner Stille
    email                : stille@uni-freiburg.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef TEXTITEM_H
#define TEXTITEM_H

#include <QGraphicsItem>
#include <QTextDocument>
#include "decoitem.h"

/**
  * Text item class. Class for text decoration items.
  * @author Werner Stille
  */
class KDE_EXPORT TextItem : public DecoItem, public QGraphicsItem
{
public:
  /** Text alignment. Determines how text will be aligned horizontally. */
  enum Alignment {
    AlignLeft = 1, ///< Aligns at left side.
    AlignCentered, ///< Aligns at center.
    AlignRight     ///< Aligns at right side.
  };
  /** Specific QGraphicsItem type number for function type() */
  enum {Type = UserType + Text
    /**< Specific QGraphicsItem type number of TextItem for function type(). */
  };
  /**
    * Constructor. Initializes corresponding to arguments.
    * @param x0 x position in cm.
    * @param y0 y position in cm.
    * @param color color for text.
    * @param parent pointer to parent item.
    * @param text text.
    * @param visible true for visible item.
    * @param alignment horizontal alignment.
    * @param relSize relative size of text.
    * @param dir direction in degree.
    */
  TextItem(double x0 = 0, double y0 = 0, unsigned color = 0,
           QGraphicsItem* parent = 0, const QString& text = "",
           bool visible = true, Alignment alignment = AlignLeft,
           double relSize = 1, double dir = 0);
  /**
    * Constructor. Initializes corresponding to plot file.
    * @param plo pointer to KSimpleConfig object containing item data.
    * @param color default color for text.
    * @param parent pointer to parent item.
    */
  TextItem(const KConfigGroup* plo, unsigned color = 0,
           QGraphicsItem* parent = 0);
  /**
    * Copy constructor.
    * @param a source item.
    */
  TextItem(const TextItem& a);
  /** Destructor. */
  virtual ~TextItem();
  /** Returns horizontal alignment. */
  inline Alignment alignment() const { return m_alignment; };
  /** Calculates bounding rectangle. */
  virtual QRectF boundingRect() const;
  /** Returns rotation angle in deg. */
  inline double dir() const { return m_dir; };
  /** Paints item. */
  virtual void paint(QPainter* painter, const QStyleOptionGraphicsItem* option,
                     QWidget* widget = 0);
  /**
    * Sets alignment.
    * @param alignment horizontal alignment.
    */
  inline void setAlignment(Alignment alignment) { m_alignment = alignment; };
  /**
    * Sets rotation angle.
    * @param dir rotation angle in deg.
    */
  inline void setDir(double dir) { m_dir = dir; };
  /**
    * Sets properties according to transformation matrix.
    * @param m transformation matrix.
    */
  void setProperties(const QTransform& m);
  /** Sets size according to transformation matrix. */
  void setSize();
  /**
    * Sets text.
    * @param text text.
    */
  inline void setText(const QString& text) { m_text = text; };
  /**
    * Sets text.
    * @param doc document.
    * @param text text.
    * @param color color.
    */
  static void setText(QTextDocument& doc, const QString& text,
                      unsigned color = 0);
  /** Returns text. */
  inline QString text() const { return m_text; };
  /** Returns specific QGraphicsItem type number. */
  virtual int type() const { return Type; };
  /** Applies changes and forces redrawing. */
  void update();
  /**
   * Writes into plot file.
   * @param plo KConfigGroup object.
   */
  void writePlo(KConfigGroup* plo) const;

protected:
  /** Performs initialization. */
  void init();
  /** Updates position. */
  virtual void mouseReleaseEvent(QGraphicsSceneMouseEvent* event);
  /** Sets properties according to transformation matrix. */
  void setProperties();
  Alignment m_alignment;
  double m_dir;
  QString m_text;
  QTextDocument m_doc;
};

#endif
