/***************************************************************************
                          fitdlg.cpp  -  description
                             -------------------
    begin                : Tue May 25 1999
    copyright            : (C) 2015 by Werner Stille
    email                : stille@uni-freiburg.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <QCheckBox>
#include <QDesktopWidget>
#include <QFileInfo>
#include <QGroupBox>
#include <QHeaderView>
#include <QLabel>
#include <QLayout>
#include <QListWidget>
#include <QTableWidget>
#include <QTimer>
#include <KApplication>
#include <KConfig>
#include <KLocalizedString>
#include <KMessageBox>
#include <KNotification>
#include <KPushButton>
#include <KUndoStack>
#include "kpl.h"
#include "kplscene.h"
#include "arrayitem.h"
#include "doubleedit.h"
#include "errmoddlg.h"
#include "fitdlg.h"
#include "frameitem.h"
#include "funitem.h"
#include "kpldoc.h"
#include "kplsettings.h"
#include "spinbox.h"
#include "lmfit.h"
#include "residualsdlg.h"
#include "sourcewidget.h"

CorrDlg::CorrDlg(QWidget* parent, int np, const int* ip, const double* corr) :
  KDialog(parent)
{
  setPlainCaption(i18n("Parameter correlation coefficients"));
  setButtons(Close);
  QTableWidget* table = new QTableWidget(np, np);
  QStringList labels;
  for (int ix = 0; ix < np; ++ix) {
    labels << QString("p%1").arg(ip[ix]);
    for (int iy = 0; iy < np; iy++) {
      double temp = corr[np * iy + ix];
      QTableWidgetItem* item = new QTableWidgetItem(((temp >= -1.0004) &&
                                                     (temp  <= 1.0004)) ?
                                                    QString::number(temp, 'f',
                                                                    3) :
                                                    "NaN");
      item->setFlags(Qt::ItemIsEnabled);
      table->setItem(iy, ix, item);
    }
  }
  table->setHorizontalHeaderLabels(labels);
  table->setVerticalHeaderLabels(labels);
  table->resizeColumnsToContents();
  int w = table->verticalHeader()->width() + 16;
  int h = table->horizontalHeader()->height() + 8;
  for (int i = 0; i < np; ++i) {
    w += table->columnWidth(i);
    h += table->rowHeight(i);
  }
  table->setMinimumSize(QSize(w, h).
    boundedTo(0.8 * QApplication::desktop()->size()));
  setMainWidget(table);
  resize(minimumSize());
}

CorrDlg::~CorrDlg()
{
}

FitDlg::FitDlg(QWidget* parent, KplDoc* model, const QList<const ArrayItem*>* ai0,
               const QList<FunItem*>& fi0, int mode) :
 KDialog(parent), running(false), dlgMode(mode), m(model), ai(ai0), fi(fi0),
 lm(0), err(0), sw(0)
{
  err = new FunItem::DataErrorStruct[ai->count()];
  FunItem::DataErrorStruct err0 = m->errorModel();
  for (int i = 0; i < ai->count(); ++i) {
    err[i] = err0;
    const ArrayItem* a = ai->at(i);
    bool validErrCol = (a->colErr() < a->numCols());
    if (dlgMode & AutoFit)
      err[i].fitErrCol = validErrCol ? err0.fitErrCol : false;
    else {
      QString groupName = QString("DataErrors%1").arg(i);
      if (KGlobal::config()->hasGroup(groupName)) {
        KConfigGroup config = KGlobal::config()->group(groupName);
        err[i].errModPath = config.readEntry("ErrorModelPath",
                                             err0.errModPath);
        err[i].errModName = config.readEntry("ErrorModelName",
                                             err0.errModName);
        FunItem::stringToPar(config.readEntry("ErrorModelParam",
                                              KplSettings::errorModelParam()),
                             err[i].parErrMod);
        err[i].fitErrCol = validErrCol ?
          config.readEntry("FitErrorColumn", err0.fitErrCol) : false;
        err[i].errModArg = FunItem::DataErrorStruct::
          ErrorModelArgument(config.readEntry("ErrorModelArgument",
                                              int(err0.errModArg)));
      }
    }
  }
  FunItem* f = fi.first();
  for (int i = 0; i < FunItem::NumParMax; ++i) {
    m->pLast()[i] = f->parY()[i];
    m->pErr()[i] = 0;
  }
  if (dlgMode & ShowDlg) {
    setCaption(i18n("Parameter fit"));
    setButtons(Help | Ok | Apply | Cancel | User1 | User2);
    setButtonText(User1, i18n("&Start"));
    setButtonText(User2, i18n("&Residuals"));
    setButtonFocus(Ok);
    QWidget* page = new QWidget;
    setMainWidget(page);
    QHBoxLayout* hbox = 0;
    QVBoxLayout* vbox;
    if (KplSettings::showSource()) {
      hbox = new QHBoxLayout(page);
      hbox->setSpacing(spacingHint());
      hbox->addLayout(vbox = new QVBoxLayout);
    } else {
      vbox = new QVBoxLayout(page);
      vbox->setSpacing(spacingHint());
    }
    QGroupBox* g = new QGroupBox(i18n("Parameter"));
    vbox->addWidget(g);
    QVBoxLayout* vbox2 = new QVBoxLayout;
    g->setLayout(vbox2);
    int nGr = 1 + (FunItem::NumParMax - 1) / 10;
    QGridLayout* grid = new QGridLayout;
    vbox2->addLayout(grid);
    for (int i = 1; i < nGr; ++i)
      grid->addItem(new QSpacerItem(20, 10), 0, 4 * i);
    char frm = m->presentation();
    int prec = KplSettings::precision();
    int w = 0;
    for (int i = 0; i < FunItem::NumParMax; ++i) {
      int ix = 5 * (i / 10);
      int iy = i % 10;
      grid->addWidget(enFit[i] = new QCheckBox(QString("p%1").arg(i)), iy, ix);
      enFit[i]->setChecked(m->fitPar()[i]);
      widgetList << enFit[i];
      grid->addWidget(par[i] = new DoubleEdit(m->pLast()[i], frm, prec), iy,
                      ix + 1);
      if (!i)
        w = qMax(80, par[0]->fontMetrics().
                 width(QString::number(-1e-123 / 3, frm, prec + 1)) + 3);
      par[i]->setFixedWidth(w);
      widgetList << par[i];
      grid->addWidget(new QLabel("\261"), iy, ix + 2);
      grid->addWidget(epar[i] = new QLabel("0"), iy, ix + 3);
      epar[i]->setFrameStyle(QFrame::Panel | QFrame::Sunken);
      epar[i]->setFixedWidth(w);
    }
    QHBoxLayout* hbox2 = new QHBoxLayout;
    vbox2->addLayout(hbox2);
    hbox2->addWidget(nonLin = new QCheckBox(i18n("Nonlinear fit")));
    nonLin->setChecked(m->fitNonLinear());
    widgetList << nonLin;
    QPushButton* b;
    hbox2->addWidget(b = new QPushButton(i18n("Load")));
    connect(b, SIGNAL(clicked()), SLOT(slotGetParameter()));
    widgetList << b;
    hbox2->addWidget(b = new QPushButton(i18n("Save")));
    connect(b, SIGNAL(clicked()), SLOT(slotSaveParameter()));
    widgetList << b;
    hbox2->addWidget(showCorr = new QPushButton(i18n("Correlations")));
    showCorr->setEnabled(false);
    connect(showCorr, SIGNAL(clicked()), SLOT(slotCorr()));
    g = new QGroupBox(i18n("Data errors"));
    g->setLayout(hbox2 = new QHBoxLayout);
    hbox2->addWidget(new QLabel(i18n("Array")));
    hbox2->addWidget(sArr = new SpinBox(0, 0, ai->count() - 1));
    hbox2->addWidget(lArr = new QLabel);
    lArr->setFrameStyle(QFrame::Panel | QFrame::Sunken);
    hbox2->addWidget(errCol = new QCheckBox(i18n("Error column")));
    hbox2->addItem(new QSpacerItem(10, 10, QSizePolicy::Expanding));
    hbox2->addWidget(errMod = new QPushButton(i18n("Model")));
    connect(errMod, SIGNAL(clicked()), SLOT(slotErrMod()));
    vbox->addWidget(g);
    vbox->addLayout(hbox2 = new QHBoxLayout);
    hbox2->addWidget(new QLabel(i18n("Maximum iterations")));
    hbox2->addWidget(sItmax = new SpinBox(m->fitMaxIterations(), 1, 200));
    widgetList << sItmax;
    hbox2->addItem(new QSpacerItem(10, 10, QSizePolicy::Expanding));
    hbox2->addWidget(new QLabel(i18n("Tolerance")));
    hbox2->addWidget(eTol = new DoubleEdit(m->fitTolerance(), 0, 1, 'g', 6));
    widgetList << eTol;
    vbox->addWidget(results = new QListWidget);
    results->setMinimumHeight(40);
    widgetList << button(Ok);
    widgetList << button(Apply);
    widgetList << button(User1);
    widgetList << button(User2);
    enableButton(User2, false);
    if (KplSettings::showSource()) {
      hbox->addWidget(sw = new SourceWidget, 1);
      sw->setMinimumWidth(200);
    }
    setHelp("SEC-FIT");
    restoreDialogSize(KGlobal::config()->group("FitDlg"));
    connect(nonLin, SIGNAL(toggled(bool)), SLOT(enableNonLin(bool)));
    connect(errCol, SIGNAL(toggled(bool)), SLOT(errColToggled(bool)));
    connect(sArr, SIGNAL(valueChanged(int)), SLOT(updArray()));
    enableNonLin(m->fitNonLinear());
    QTimer::singleShot(0, this, SLOT(updArray()));
  }
}

FitDlg::~FitDlg()
{
  if (dlgMode & ShowDlg) {
    KConfigGroup config = KGlobal::config()->group("FitDlg");
    saveDialogSize(config);
  }
  delete [] err;
}

void FitDlg::apply()
{
  slotButtonClicked(Apply);
}

void FitDlg::fit()
{
  np = 0;
  for (int i = 0; i < FunItem::NumParMax; ++i) {
    if (dlgMode & ShowDlg)
      m->fitPar()[i] = enFit[i]->isChecked();
    if (m->fitPar()[i])
      ++np;
  }
  if (!np) {
    KMessageBox::sorry(this, i18n("No parameter to fit!"));
    return;
  }
  lm = new LMFit(ai, &fi, m->fitPar(), m->pLast(), this, sig.data());
  bool nLin = m->fitNonLinear();
  double tol = m->fitTolerance();
  int itMax = m->fitMaxIterations();
  if (dlgMode & ShowDlg) {
    nLin = nonLin->isChecked();
    if (nLin) {
      tol = eTol->value();
      itMax = sItmax->value();
    }
    showCorr->setEnabled(false);
    results->clear();
    connect(lm, SIGNAL(updateMessage(const QString &)),
            SLOT(slotMessage(const QString &)));
    for (int i = 0; i < FunItem::NumParMax; ++i)
      m->pLast()[i] = par[i]->value();
    foreach (QWidget* w, widgetList)
      w->setEnabled(false);
  }
  kapp->processEvents();
  double avgErr;
  running = true;
  int info = FunItem::fit(ai, &fi, m->fitPar(), m->pLast(), &fvec, &sig, err,
                          nLin, tol, itMax, lm, &m->chiSquare(), corr,
                          m->pErr(), &avgErr, this);
  if (info == -101)
    KMessageBox::error(this, i18n("no function address!"));
  QString s;
  if (info >= -100) {
    if (nLin) {
      if (info >= 8)
        info = 4;
      if (dlgMode & ShowDlg) {
        switch (info) {
          case 0:
            s = i18n("Improper input parameters");
            break;
          case 1:
            s = i18n("Relative error in chi-square is at most tolerance");
            break;
          case 2:
            s = i18n("Relative error in parameters is at most tolerance");
            break;
          case 3:
            s = i18n("Relative error in chi-square and in parameters is at "
                     "most tolerance");
            break;
          case 4:
            s = i18n("Function vector is orthogonal to the columns of "
                     "the jacobian");
            break;
          case 5:
            s = i18n("Number of iterations has reached or exceeded maximum");
            break;
          case 6:
            s = i18n("Tolerance too small. No further reduction of "
                     "chi-square possible");
            break;
          case 7:
            s = i18n("Tolerance too small. No further improvement of "
                     "parameters possible");
            break;
          default:
            s = i18n("Terminated by user");
        }
        slotMessage(s);
      }
    }
  }
  running = false;
  if (info > -100) {
    char frm = m->presentation();
    int prec = KplSettings::precision();
    int k = 0;
    for (int i = 0; i < FunItem::NumParMax; ++i) {
      if (m->fitPar()[i])
        ip[k++] = i;
      if (dlgMode & ShowDlg) {
        par[i]->setValue(m->pLast()[i]);
        epar[i]->setText(QString::number(m->pErr()[i], frm, prec));
      }
      if (dlgMode & Follow)
        m->pFit()[i] = m->pLast()[i];
    }
    if (dlgMode & ShowDlg) {
      int n = 0;
      foreach (const ArrayItem* a, *ai)
        n += a->count();
      int ny = n - np;
      double q = ny ? LMFit::igamc(0.5 * ny, 0.5 * m->chiSquare()) : 0;
      slotMessage(QString(i18n("Average error = %1    ny = %2    "
                               "Significance Q = %3")).arg(avgErr, 0, 'g', 3).
                  arg(ny).arg(q, 0, 'g', 3));
      showCorr->setEnabled(true);
    }
    if (dlgMode & SavePar) {
      KUrl url = m->url();
      QFileInfo inf(url.path());
      url.setPath(inf.absolutePath() + "/" + inf.baseName() + ".par");
      m->saveParameter(this, url, m->pLast(), m->pErr());
    }
    if (!(dlgMode & ShowDlg))
      slotButtonClicked(Apply);
    if (!lm->userBreak())
      KNotification::event("fitCompleted", i18n("Parameter fit completed."));
  }
  delete lm;
  if (dlgMode & ShowDlg) {
    foreach (QWidget* w, widgetList)
      w->setEnabled(true);
    enableNonLin(nonLin->isChecked());
    QTimer::singleShot(0, results, SLOT(scrollToBottom()));
  }
}

void FitDlg::errColToggled(bool state)
{
  errMod->setEnabled(!state);
  err[sArr->value()].fitErrCol = state;
}

void FitDlg::enableNonLin(bool on)
{
  eTol->setEnabled(on);
  sItmax->setEnabled(on);
}

void FitDlg::slotButtonClicked(int button)
{
  switch (button) {
    case Apply:
    case Ok: {
      if (dlgMode & ShowDlg) {
        for (int i = 0; i < FunItem::NumParMax; ++i)
          m->fitPar()[i] = enFit[i]->isChecked();
        m->fitNonLinear() = nonLin->isChecked();
        if (m->fitNonLinear()) {
          m->fitMaxIterations() = sItmax->value();
          m->fitTolerance() = eTol->value();
        }
        char frm = m->presentation();
        int prec = KplSettings::precision();
        for (int j = 0; j < ai->count(); ++j) {
          KConfigGroup config = KGlobal::config()->
            group(QString("DataErrors%1").arg(j));
          config.writeEntry("FitErrorColumn", err[j].fitErrCol);
          config.writeEntry("ErrorModelPath", err[j].errModPath);
          config.writeEntry("ErrorModelName", err[j].errModName);
          config.writeEntry("ErrorModelParam",
                            FunItem::parToString(err[j].parErrMod, frm, prec));
          config.writeEntry("ErrorModelArgument", int(err[j].errModArg));
        }
      }
      QUndoCommand* macro = new QUndoCommand;
      QList<FunItem*> newItems;
      foreach (FunItem* item, fi) {
        FunItem* newItem = new FunItem(*item);
        for (int i = 0; i < FunItem::NumParMax; ++i)
          newItem->setPar(i, m->pLast()[i]);
        newItems << newItem;
        new EditCommand(newItem, item, m, macro);
      }
      m->undoStack()->push(macro);
      if (button == Ok)
        accept();
      else
        fi = newItems;
      break;
    }
    case Cancel:
      if (running)
        lm->setUserBreak();
      else
        reject();
      break;
    case User1:
      fit();
      break;
    case User2: {
      KplScene scene;
      FrameItem fri(4, 3, 15, 10, 0, 0, false, false,
                    FrameItem::GridWithLabels, 15, 10, "x", i18n("Residuals"),
                    "", 0.01 * KplSettings::relSize());
      scene.insertItem(&fri);
      int j = 0;
      foreach (const ArrayItem* a, *ai) {
        ArrayItem* ai2 = new ArrayItem(a->color(), 0, 1, 2, true, a->symbol(),
          ArrayItem::Point, KUrl(), 1, 1, 0, a->count(), 0,
          KplItem::NoBrush, 0, 0, true, true);
        double** x0 = a->data();
        double** x = ai2->data();
        int ix = a->colX();
        int i0 = a->rowStart();
        for (int i = 0; i < a->count(); ++i) {
          x[0][i] = x0[ix][i0 + i];
          x[1][i] = fvec[j] * sig[j];
          x[2][i] = sig[j++];
        }
        scene.insertItem(ai2);
      }
      fri.autoScale();
      ResidualsDlg dlg(&scene, this);
      dlg.exec();
      break;
    }
    default:
      KDialog::slotButtonClicked(button);
  }
}

void FitDlg::slotCorr()
{
  CorrDlg dlg(this, np, ip, corr);
  dlg.exec();
}

void FitDlg::slotErrMod()
{
  ErrModDlg dlg(this, m, &err[sArr->value()]);
  dlg.exec();
}

void FitDlg::slotGetParameter()
{
  if (m->getPar(this, m->pLast()))
    for (int i = 0; i < FunItem::NumParMax; ++i) {
      par[i]->setValue(m->pLast()[i]);
      m->pErr()[i] = 0;
      epar[i]->setText("0");
    }
}

void FitDlg::slotMessage(const QString& msg)
{
  new QListWidgetItem(msg, results);
  results->scrollToBottom();
  kapp->processEvents();
}

void FitDlg::slotSaveParameter()
{
  KUrl url;
  if (Kpl::getWriteUrl(this, url, KplSettings::currentDir(), "*.par\n*"))
    m->saveParameter(this, url, m->pLast(), m->pErr());
}

void FitDlg::updArray()
{
  int i = sArr->value();
  const ArrayItem* a = ai->at(i);
  lArr->setText(a->url().fileName() + QString(" %1 %2 %3").arg(a->colX()).
    arg(a->colY()).arg(a->colErr()));
  errCol->setChecked(err[i].fitErrCol);
  errCol->setEnabled(a->colErr() < a->numCols());
  if (KplSettings::showSource()) {
    FunItem* f = fi.at(i);
    QFileInfo info(f->urlY().path());
    QString s = info.absolutePath() + "/" + info.baseName();
    QFile file(s + ".c");
    bool open = file.open(QIODevice::ReadOnly);
    if (!open) {
      file.setFileName(s + ".cpp");
      open = file.open(QIODevice::ReadOnly);
    }
    if (open) {
      sw->fill(&file);
      sw->highlight(f->nameY());
    }
  }
}
