/***************************************************************************
                          kpldoc.h  -  description
                             -------------------

    This file is a part of kpl - a program for graphical presentation of
    data sets and functions.

    begin                : Sat Apr 24 15:14:00 MEST 1999

    copyright            : (C) 2015 by Werner Stille
    email                : stille@uni-freiburg.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef KPLDOC_H
#define KPLDOC_H

#include <QAbstractItemModel>
#include <QDateTime>
#include <QList>
#include <QPointF>
#include <QString>
#include <QUndoCommand>
#include <KUndoStack>
#include "arrayitem.h"
#include "funitem.h"
#include "kplscene.h"

class QGraphicsItem;
class QLineEdit;
class KplDoc;
class FrameItem;

/**
  * Delete command class.
  * Provides methods for undo/redo of delete commands.
  * @author Werner Stille
  */
class DeleteCommand : public QUndoCommand
{
public:
  /**
    * Constructor.
    * @param it pointer to plot item.
    * @param d pointer to model.
    * @param parent pointer to parent command.
    */
  DeleteCommand(QGraphicsItem* it, KplDoc* d, QUndoCommand* parent = 0);
  /** Destructor. */
  virtual ~DeleteCommand();
  /** Reimplemented from QUndoCommand. */
  virtual void redo();
  /** Reimplemented from QUndoCommand. */
  virtual void undo();

protected:
  QGraphicsItem *item, *parentItem, *newFrame;
  QList<QGraphicsItem*> childItems;
  KplDoc* doc;
};

/**
  * Edit command class.
  * Provides methods for undo/redo of edit commands.
  * @author Werner Stille
  */
class EditCommand : public QUndoCommand
{
public:
  /**
    * Constructor.
    * @param it pointer to edited plot item.
    * @param old pointer to old plot item.
    * @param d pointer to model.
    * @param parent pointer to parent command.
    */
  EditCommand(QGraphicsItem* it, QGraphicsItem* old, KplDoc* d,
              QUndoCommand* parent = 0);
  /** Destructor. */
  virtual ~EditCommand();
  /** Reimplemented from QUndoCommand. */
  virtual void redo();
  /** Reimplemented from QUndoCommand. */
  virtual void undo();

protected:
  void swapItems(QGraphicsItem* n, QGraphicsItem* o);
  QGraphicsItem *item, *oldItem, *parentItem;
  KplDoc* doc;
};

/**
  * Move command class.
  * Provides methods for undo/redo of move commands.
  * @author Werner Stille
  */
class MoveCommand : public QUndoCommand
{
public:
  /**
    * Constructor.
    * @param iSource source item list index.
    * @param iDest destination item list index.
    * @param d pointer to model.
    */
  MoveCommand(int iSource, int iDest, KplDoc* d);
  /** Destructor. */
  virtual ~MoveCommand();
  /** Reimplemented from QUndoCommand. */
  virtual void redo();
  /** Reimplemented from QUndoCommand. */
  virtual void undo();

protected:
  void swapItems(int i1, int i2);
  int is, id;
  KplDoc* doc;
};

/**
  * New command class.
  * Provides methods for undo/redo of new commands.
  * @author Werner Stille
  */
class NewCommand : public QUndoCommand
{
public:
  /**
    * Constructor.
    * @param it pointer to plot item.
    * @param d pointer to model.
    * @param parent pointer to parent command.
    */
  NewCommand(QGraphicsItem* it, KplDoc* d, QUndoCommand* parent = 0);
  /** Destructor. */
  virtual ~NewCommand();
  /** Reimplemented from QUndoCommand. */
  virtual void redo();
  /** Reimplemented from QUndoCommand. */
  virtual void undo();

protected:
  QGraphicsItem* item, *parentItem;
  QList<QGraphicsItem*> childItems, oldParents;
  KplDoc* doc;
};

/**
  * Toggle active command class.
  * Provides methods for undo/redo of toggle visibility commands.
  * @author Werner Stille
  */
class ToggleActiveCommand : public QUndoCommand
{
public:
  /**
    * Constructor.
    * @param it pointer to plot item.
    * @param d pointer to model.
    */
  ToggleActiveCommand(QGraphicsItem* it, KplDoc* d);
  /** Destructor. */
  virtual ~ToggleActiveCommand();
  /** Reimplemented from QUndoCommand. */
  virtual void redo();
  /** Reimplemented from QUndoCommand. */
  virtual void undo();

protected:
  QGraphicsItem* item, *newFrame;
  QList<QGraphicsItem*> childItems;
  KplDoc* doc;
};

/**
  * Toggle collapsed command class.
  * Provides methods for undo/redo of toggle collapsed commands.
  * @author Werner Stille
  */
class ToggleCollapsedCommand : public QUndoCommand
{
public:
  /**
    * Constructor.
    * @param it pointer to plot item.
    * @param d pointer to model.
    */
  ToggleCollapsedCommand(FrameItem* it, KplDoc* d);
  /** Destructor. */
  virtual ~ToggleCollapsedCommand();
  /** Reimplemented from QUndoCommand. */
  virtual void redo();
  /** Reimplemented from QUndoCommand. */
  virtual void undo();

protected:
  FrameItem* item;
  KplDoc* doc;
};

/**
  * Kpl model class.
  * Provides methods for reading, writing and scaling of plot data.
  * @author Werner Stille
  */
class KplDoc : public QAbstractItemModel
{
  Q_OBJECT
public:
  /**
   * Constructor.
   * @param parent pointer to parent QObject.
   */
  KplDoc(QObject* parent = 0);
  /** Destructor. */
  virtual ~KplDoc();
  /** Types of files */
  enum FileType{Unknown, Data, Plot};
  /** Returns chi square value of last fit. */
  inline double& chiSquare() { return m_chiSquare; };
  /** Reimplemented from QAbstractItemModel. */
  virtual int columnCount(const QModelIndex& parent = QModelIndex()) const;
  /**
   * Copies plot items to clipboard.
   * @param indexes list of item model indexes.
   * @param cut true to delete from source list.
   */
  void copyItems(QModelIndexList indexes, bool cut = false);
  /** Reimplemented from QAbstractItemModel. */
  virtual QVariant data(const QModelIndex& index,
                        int role = Qt::DisplayRole) const;
  /** Reimplemented from QAbstractItemModel. */
  virtual bool dropMimeData(const QMimeData* data, Qt::DropAction action,
                            int row, int column, const QModelIndex& parent);
  /**
   * Opens dialog to edit plot item.
   * @param indexes list of item model indexes.
   * @param pos position for dialog window.
   */
  void editItem(QModelIndexList indexes, QPoint pos);
  /** Returns structure for parameter fitting error model. */
  const FunItem::DataErrorStruct& errorModel();
  /**
   * Returns the type of current document file.
   * @return file type (Plot or Data).
   */
  inline FileType fileType() const { return m_fileType; };
  /**
   * Performs fitting.
   * @param indexes list of item model indexes.
   */
  void fitItems(QModelIndexList indexes);
  /** Returns maximum number of parameter fit iterations. */
  inline int& fitMaxIterations() { return m_fitMaxIterations; };
  /** Returns true for nonlinear parameter fitting. */
  inline bool& fitNonLinear() { return m_fitNonLinear; };
  /** Returns array with boolean entries for parameters to fit. */
  inline bool* fitPar() { return m_fitPar; };
  /**
   * Returns true for fittable plot item list.
   * @param indexes list of item model indexes.
   */
  bool fittable(QModelIndexList indexes);
  /** Returns relative fit tolerance value for parameter fits. */
  inline double& fitTolerance() { return m_fitTolerance; };
  /** Reimplemented from QAbstractItemModel. */
  virtual Qt::ItemFlags flags(const QModelIndex& index) const;
  /**
   * Opens a dialog to select a function parameter file.
   * @param parent parent widget.
   * @param p double array for parameters.
   * @return true for success.
   */
  bool getPar(QWidget* parent, double* p);
  /** Reimplemented from QAbstractItemModel. */
  virtual QVariant headerData(int section, Qt::Orientation orientation,
                              int role = Qt::DisplayRole) const;
  /** Reimplemented from QAbstractItemModel. */
  virtual QModelIndex index(int row, int column,
                            const QModelIndex& parent = QModelIndex()) const;
  /**
   * Inserts plot item into item list.
   * @param item pointer to item.
   * @param z z value of item, -1 for appending.
   */
  void insertItem(QGraphicsItem* item, int z = -1);
  /**
   * Inserts plot items from plot file into item list.
   * @param url URL of plot file.
   * @param z z value of first item to insert, -1 for appending.
   */
  void insertItems(const KUrl& url, int z);
  /**
   * Loads a function parameter file.
   * @param url file URL.
   * @param parent parent widget.
   * @param p double array for parameters.
   * @param idec decimal character: Point or Comma.
   */
  bool loadPar(const KUrl& url, QWidget* parent, double* p,
               ArrayItem::DecimalCharacter idec);
  /** Reimplemented from QAbstractItemModel. */
  virtual QMimeData* mimeData(const QModelIndexList& indexes) const;
  /** Reimplemented from QAbstractItemModel. */
  virtual QStringList mimeTypes() const;
  /** Initializes the document generally. */
  void newDocument();
  /**
   * Inserts new plot item and opens dialog to edit it.
   * @param type item type.
   * @param z z value of item, -1 for appending.
   * @param pos position for dialog window.
   */
  void newItem(int type, int z, const QPoint& pos);
  /**
   * Reads file and initializes plot items.
   * @param url URL of the file.
   * @param idec decimal character: Point or Comma.
   * @param type one of the following: Unknown, Data, Plot.
   * @return true for success.
   */
  bool openDocument(const KUrl& url, ArrayItem::DecimalCharacter idec,
                    FileType type);
  /** Reimplemented from QAbstractItemModel. */
  virtual QModelIndex parent(const QModelIndex& index) const;
  /** Returns array with parameter errors of last fit. */
  double* pErr() { return m_pErr; };
  /** Returns array with parameters. */
  double* pFit() { return m_pFit; };
  /** Returns array with parameters of last fit. */
  double* pLast() { return m_pLast; };
  /** Returns C format character for presentation of double values. */
  static char presentation();
  /**
   * Sets array with boolean entries for parameters to fit corresponding
   * to Kpl's settings.
   * @param fitPar boolean array.
   */
  static void readFitPar(bool* fitPar);
  /**
   * Removes plot items from list.
   * @param indexes list of item model indexes.
   */
  void removeItems(QModelIndexList indexes);
  /** Reimplemented from QAbstractItemModel. */
  void reset();
  /** Reimplemented from QAbstractItemModel. */
  virtual int rowCount(const QModelIndex& parent = QModelIndex()) const;
  /**
   * Saves the document under URL.
   * @param url URL of file.
   * @param absolute save absolute URL when true.
   * @return true for success.
   */
  bool saveDocument(const KUrl& url, bool absolute = true);
  /**
   * Opens a dialog to save function parameters in a file.
   * @param parent parent widget.
   * @param p double array with parameters.
   */
  void saveFunPar(QWidget* parent, const double* p);
  /**
   * Saves parameters in a file.
   * @param parent parent widget.
   * @param url file URL.
   * @param p double array with parameters.
   * @param err double array with parameters (0 for no error output).
   * @return true for success.
   */
  bool saveParameter(QWidget* parent, const KUrl& url, const double* p,
                     const double* err);
  /** Returns pointer to Kpl's graphics scene. */
  inline KplScene* scene() { return m_scene; };
  static char separator();
  /**
   * Sets the current directory path.
   * @param url current path.
   */
  void setCurrentDir(const KUrl& url);
  /** Reimplemented from QAbstractItemModel. */
  virtual bool setData(const QModelIndex& index, const QVariant& value,
                       int role = Qt::EditRole);
  /**
   * Sets structure for parameter fit error model.
   * @param err error model structure containing the values to be set.
   */
  inline void setErrorModel(const FunItem::DataErrorStruct& err)
    { m_errorModel = err; };
  /**
   * Set the type of current document file.
   * @param type file type (Plot or Data).
   */
  inline void setFileType(FileType type) { m_fileType = type; };
  /**
   * Returns settings text of item.
   * @param item pointer to item.
   */
  static QString settings(QGraphicsItem* item);
  /**
   * Sets the URL of the document.
   * @param url URL.
   */
  void setUrl(const KUrl& url);
  /** Reimplemented from QAbstractItemModel. */
  virtual Qt::DropActions supportedDropActions() const;
  /**
   * Swaps parents of plot items.
   * @param n pointer to new parent item.
   * @param o pointer to old parent item.
   */
  void swapParents(QGraphicsItem* n, QGraphicsItem* o) const;
  /**
   * Returns item type name.
   * @param item pointer to item.
   */
  static const QString& typeName(const QGraphicsItem* item);
  /** Returns pointer to stack of undo commands. */
  inline KUndoStack* undoStack() const { return m_undoStack; };
  /**
   * Returns the URL of the document.
   * @return URL.
   */
  inline const KUrl& url() const { return m_url; };
  /**
   * Returns the modification time of the URL of the document.
   * @return modification date and time.
   */
  inline const QDateTime& urlTime() const { return m_time; };

public slots:
  /**
   * Collapses frame item in item list.
   * @param index model index of frame item.
   */
  void collapse(const QModelIndex& index);
  /**
   * Expands frame item in item list.
   * @param index model index of frame item.
   */
  void expand(const QModelIndex& index);
  /**
   * Zooms plot.
   * @param rect new rectangular plot range in pixel units.
   */
  void zoom(const QRectF& rect);

signals:
  /**
   * Emitted when status message should be displayed.
   * @param message message text.
   */
  void displayMessage(const QString& message);
  /**
   * Emitted when frame item is expanded.
   * @param index model index of frame item.
   */
  void expanded(const QModelIndex& index);

protected:
  /**
   * Reads plot file and initializes plot items.
   * @param url URL of the file.
   * @param list list for plot items.
   * @return true for success.
   */
  bool readPlo(const KUrl& url, QList<QGraphicsItem*>& list) const;
  /**
   * Sets collapsed state of frame item.
   * @param index model index of frame item.
   * @param state true for collapsed.
   */
  void setCollapsed(const QModelIndex& index, bool state);
  /**
   * Updates position of plot item.
   * @param item pointer to item.
   * @param p position of bottom left corner in cm.
   * @param macro pointer to parent command.
   */
  void updatePos(QGraphicsItem* item, const QPointF& p,
                 QUndoCommand* macro = 0);
  /**
   * Updates position of plot item.
   * @param item pointer to item.
   * @param p position of bottom left corner in cm.
   * @param macro pointer to parent command.
   */
  template<typename T> void updatePos2(T* item, const QPointF& p,
                                       QUndoCommand* macro)
  {
    T* o = qgraphicsitem_cast<T*>(item);
    if (macro) {
      T* n = new T(*o);
      n->updatePos(p, false);
      new EditCommand(n, o, this, macro);
    } else {
      QPointF oldPos(o->x0(), o->y0());
      if (oldPos == p)
        return;
      T* n = new T(*o);
      n->updatePos(p);
      QList<QGraphicsItem*> list = m_scene->selectedItems();
      if (list.count() > 1) {
        QPointF delta = QPointF(n->x0(), n->y0()) - oldPos;
        macro = new QUndoCommand;
        foreach (QGraphicsItem* item2, list)
          if (item2 != item)
            updatePos(item2, delta, macro);
      }
      EditCommand* cmd = new EditCommand(n, o, this, macro);
      m_undoStack->push(macro ? macro : cmd);
    }
  }
  /**
   * Updates position of frame-like plot item.
   * @param item pointer to item.
   */
  template<typename T> void updatePos3(T* item)
  {
    QList<QGraphicsItem*> childItems = item->childItems();
    T* o = qgraphicsitem_cast<T*>(item);
    QPointF p2 = o->rectPos();
    QPointF newPos = Snap * QPointF((m_scene->mapInv(m_scene->map(p2) +
      item->pos()) / Snap).toPoint());
    if (newPos != p2) {
      T* n = new T(*o);
      n->updatePos(newPos);
      p2 = newPos - p2;
      o->setPos(0, 0);
      QUndoCommand* macro = new QUndoCommand;
      foreach (QGraphicsItem* child, childItems)
        updatePos(child, p2, macro);
      new EditCommand(n, o, this, macro);
      m_undoStack->push(macro);
    }
  }
  bool defSel, m_fitNonLinear;
  int nyLeg, m_fitMaxIterations;
  static const double Snap = 0.05;
  KplScene* m_scene;
  QWidget* app;
  FileType m_fileType;
  KUndoStack* m_undoStack;
  KUrl m_url;
  QDateTime m_time;
  mutable QModelIndexList mimeDataIndexes;
  bool m_fitPar[FunItem::NumParMax];
  double m_pFit[FunItem::NumParMax], m_pLast[FunItem::NumParMax],
         m_pErr[FunItem::NumParMax];
  double m_chiSquare, m_fitTolerance;
  FunItem::DataErrorStruct m_errorModel;

protected slots:
  /** Posts an event for plot complete notification. */
  void notify();
  /**
   * Updates position of plot item.
   * @param item pointer to item.
   */
  void updatePos(QGraphicsItem* item);
};

#endif // KPLDOC_H
