//
// C++ Interface: threeddlg
//
// Description: 
//
//
// Author: Werner Stille <stille@uni-freiburg.de>, (C) 2015
//
// Copyright: See COPYING file that comes with this distribution
//
//
#ifndef THREEDDLG_H
#define THREEDDLG_H

#include <KPageDialog>
#include "arrayitem.h"

class QCheckBox;
class QDoubleSpinBox;
class QGridLayout;
class QGroupBox;
class QLabel;
class QLineEdit;
class QPushButton;
class QSpinBox;
class QToolButton;
class QVBoxLayout;
class KButtonGroup;
class KColorButton;
class KUrlRequester;
class Array3DItem;
class DoubleEdit;
class Fun3DItem;
class KplDoc;
class SpinBox;
class ThreeDItem;

/**
  * 3D item dialog class. Provides a base dialog for the settings
  * dialogs for 3D array and 3D function items.
  * @author Werner Stille
  */
class ThreeDDlg : public KPageDialog
{
  Q_OBJECT
public:
  /**
   * Constructor.
   * @param parent pointer to parent widget.
   * @param model pointer to KplDoc model object.
   * @param ti pointer to text item.
   * @param pos dialog position.
   * @param caption window caption.
   * @param anchor context help anchor.
   * @param page1 first tab page.
   */
  ThreeDDlg(QWidget* parent, KplDoc* model, ThreeDItem* ti, const QPoint& pos,
            const QString& caption, const QString anchor,
            const QString& page1);
  /** Destructor. */
  ~ThreeDDlg();

protected:
  /**
   * Performs automatic scaling of x and z axis.
   * @param xmin minimum x value.
   * @param xmax maximum x value.
   * @param zmin minimum z value.
   * @param zmax maximum z value.
   */
  void autoScaleXZ(double xmin, double xmax, double zmin, double zmax);
  /**
   * Performs automatic scaling of y axis.
   * @param ymin minimum y value.
   * @param ymax maximum y value.
   */
  void autoScaleY(double ymin, double ymax);
  /**
    * Copies settings to 3D item object.
    * @param ti pointer to item.
    */
  void getValues(ThreeDItem* ti);
  /**
   * Initializes range widgets.
   * @param grid pointer to grid layout of dialog page.
   * @param row row index for x range widgets.
   */
  void initRange(QGridLayout* grid, int row = 0);
  KplDoc* m;
  ThreeDItem* td;
  QWidget* page1;
  QVBoxLayout* vbox1;
  QGroupBox* g1;
  DoubleEdit *exMin, *exMax, *exStep, *eyMin, *eyMax, *ezMin, *ezMax, *ezStep,
             *efx, *efy, *efz, *exTic, *eyTic, *ezTic;
  QDoubleSpinBox *ex0, *exl, *ey0, *eyl, *eXref, *eYref, *eBoxx, *eBoxy,
                 *eBoxz, *ePhi, *eTheta;
  SpinBox *emTicx, *emTicy, *emTicz, *enDigx, *enDigy, *enDigz, *ex, *ey,
          *ez, *eRelSiz;
  KButtonGroup *bg1, *bg2;
  KColorButton *colFrame, *colGrid, *colData;
  QCheckBox *modeFrame, *logX, *logY, *logZ, *autoNorm;
  QLineEdit *eLetX, *eLetY, *eLetZ, *eLetH;
  QPushButton* bAuto;

protected slots:
  /** Enables buttons. */
  void slotEdited();
  /** Displays dialog to edit the x axis label. */
  void slotEditLetX();
  /** Displays dialog to edit the y axis label. */
  void slotEditLetY();
  /** Displays dialog to edit the z axis label. */
  void slotEditLetZ();
  /** Displays dialog to edit the heading. */
  void slotEditLetH();
  /**
   * Updates dialog according to logarithmic axis option.
   * @param state state of log checkbox.
   */
  void slotLogToggled(bool state);
  /**
   * Updates dialog according to autonormalization option.
   * @param state state of autonormalization checkbox.
   */
  void slotNormToggled(bool state);
};

/**
  * 3D function settings dialog class. Provides a dialog for the settings
  * dialog of a 3D function.
  * @author Werner Stille
  */
class Fun3DDlg : public ThreeDDlg
{
  Q_OBJECT
public:
  /**
   * Constructor.
   * @param parent pointer to parent widget.
   * @param model pointer to KplDoc model object.
   * @param fi pointer to 3D function object.
   * @param pos dialog position.
   */
  Fun3DDlg(QWidget* parent, KplDoc* model, Fun3DItem *fi, const QPoint& pos);
  /** Destructor. */
  ~Fun3DDlg();

protected:
  /**
   * Copies settings to temporary function object.
   * @return false for error.
   */
  bool getTemp();
  /** Sets values of temporary function object. */
  void setTemp();
  KUrlRequester* fileName;
  QLineEdit* func;
  Fun3DItem *fi, *fit;

protected slots:
  /** Applies settings without leaving the dialog. */
  void slotApply();
  /** Performs automatic scaling. */
  void slotAutoScale();
  /** Reimplemented from KDialog. */
  virtual void slotButtonClicked(int button);
  /** Displays the dialog for parameter editing. */
  void slotEditParameter();
  /** Displays the dialog for selection of a function. */
  void slotFunc();
  /** Displays the dialog for selection of a parameter file to read. */
  void slotGetParameter();
  /** Displays the dialog for selection of a parameter file to save. */
  void slotSaveParameter();
};

/**
  * 3D array settings dialog class. Provides a dialog for the settings
  * dialog of a 3D array.
  * @author Werner Stille
  */
class Array3DDlg : public ThreeDDlg
{
  Q_OBJECT
public:
  /**
   * Constructor.
   * @param parent pointer to parent widget.
   * @param model pointer to KplDoc model object.
   * @param ai pointer to 3D array object.
   * @param pos dialog position.
   */
  Array3DDlg(QWidget* parent, KplDoc* model, Array3DItem *ai,
             const QPoint& pos);
  /** Destructor. */
  ~Array3DDlg();

protected:
  /**
   * Loads file.
   * @param url KURL of file.
   * @param idec decimal character: Point or Comma.
   * @return true when when succeeded.
   */
  bool load(KUrl &url, ArrayItem::DecimalCharacter idec);
  /**
   * Reads edited data into 3D array object.
   * @param fromBuffer read editor buffer instead of file when true.
   * @return true when when succeeded.
   */
  bool readData(bool fromBuffer = false);
  /**
   * Updates ranges of spin boxes.
   * @param ad pointer to 3D array object.
   */
  void setSpinBoxes(Array3DItem* ad);
  QString buf;
  QLabel* fileName;
  Array3DItem *ai, *ait;
  QToolButton *bReload;
  QCheckBox *err, *internal;
  QPushButton* edit;
  QSpinBox *ix, *iy, *iz, *ie, *iStart, *n;
  DoubleEdit *eFactor;

protected slots:
  /** Performs automatic scaling. */
  void slotAutoScale();
  /** Reimplemented from KDialog. */
  virtual void slotButtonClicked(int button);
  /** Edits internal data. */
  void slotEdit();
  /** Applies edited array data. */
  void slotEditApply();
  /** Displays the dialog for selection of a data file. */
  void slotFile();
  /** Toggles internal/external data. */
  void slotInternal(bool);
  /**
   * Adjusts the Start index spinbox when number of points is changed.
   * @param na actual number of points.
   */
  void slotNChanged(int na);
  /** Reloads file. */
  void slotReload();
  /**
   * Adjusts the Number of points spinbox when start index is changed.
   * @param ia actual start index.
   */
  void slotStartChanged(int ia);
};

#endif
